#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import json


class JsonObject(object):
    """
    Base JSON for all data objects.  Allows for converting JSON data into the object models.
    """

    # _json_data = False

    def __init__(self, *args, **kwargs):
        """
        If parameter values in args, then the value is expected to be a dictionary from a json response
        from the server.

        If parameter values are in kwargs, then they are named parameters passed when the object is instantiated.
        """
        if args is not None and len(args) is not 0 and args[0] is not None:
            # self._json_data = True
            for key, value in args[0].items():
                self.__dict__[key] = value
                # print('{0} : {1}'.format(key, value))

        else:
            for key, value in kwargs.items():
                self.__dict__[key] = value
                # print('{0} : {1}'.format(key, value))

    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, indent=3)

    def to_dict(self):
        data = dict()
        for key, value in self.__dict__.items():
            if not key.startswith("__") and value is not None:
                data[key] = value
        return data

    def dict_to_obj_array(self, cls, data):
        """
        # Convert dict to array of objects of type cls
        """
        if data is None:
            return None

        ol = list()
        for x in range(0, len(data)):
            ol.append(cls(data[x]))

        return ol



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os

from silentdune_client.modules.firewall.manager.slots import Slots
from silentdune_client.modules.automation.auto_discovery.base_discovery_service import BaseDiscoveryService
from silentdune_client.modules.firewall.manager.iptables_utils import create_iptables_tcp_egress_ingress_rule, create_iptables_udp_egress_ingress_rule

_logger = logging.getLogger('sd-client')


class DynamicNameServiceDiscovery(BaseDiscoveryService):
    """
    Auto discover DNS servers from local system configuration files.
    """

    _slot = Slots.dns  # DNS slot value = 130
    _config_property_name = '_disable_auto_dns'

    def _discover_iptables(self):
        """
        Look at /etc/resolv.conf file to get external DNS servers and set outbound rule(s) to allow access.
        """

        if not os.path.exists('/etc/resolv.conf'):
            _logger.error('{0}: resolv.conf not found.'.format(self.get_name()))
            return None

        rules = list()
        ipaddrs = list()

        # Get all nameserver ip address values
        with open('/etc/resolv.conf') as handle:
            for line in handle:
                if 'nameserver' in line.lower() and not line.strip().startswith('#'):
                    ipaddrs.append(line.split()[1])

        if len(ipaddrs) == 0:
            _logger.error('{0}: no name server values found in resolv.conf'.format(self.get_name()))
            return None

        for ipaddr in ipaddrs:
            try:
                _logger.debug('{0}: adding DNS IP address {1}'.format(self.get_name(), ipaddr))
                rules.append(create_iptables_tcp_egress_ingress_rule(ipaddr, 53, self._slot))
                rules.append(create_iptables_udp_egress_ingress_rule(ipaddr, 53, self._slot))
            except ValueError:
                _logger.error('{0}: Unable to validate DNS ip address {1}.'.format(self.get_name(), ipaddr))

        return rules



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from silentdune_client.models.iptables_rules import IPJumpOptions, IPJump, IPMatchOptions, \
    IPMatch, IPRule, IPRing, IPChain, IPMachineSubset

TRANSPORT_IPV4 = u'ipv4'
TRANSPORT_IPV6 = u'ipv6'
TRANSPORT_AUTO = u'auto'


def get_machine_subset(name, slot, chains, platform=u'iptables', _id=0):
    """
    Return an IPMachineSubset object
    :param name: Name for this group of rules
    :param slot: Slot ID for this group
    :param chains: List of IPChains objects
    :param platform: Must be 'iptables'
    :param _id: Id for this object
    :return:
    """
    obj = IPMachineSubset({u'name': name, u'slot': slot, u'platform': platform, u'id': _id})
    obj.chains = chains
    return obj


def get_chain(name, rings, _id=0):
    """

    :param name: iptables table name, must be 'filter', 'nat', 'mangle', 'raw' or 'security'
    :param rings: List of IPRing objects
    :param _id: Id for this object
    :return:
    """
    obj = IPChain({u'name': name, u'id': _id})
    obj.rings = rings
    return obj


def get_ring(name, version, rules, _id=0):
    """

    :param name: iptables chain name, must be 'INPUT', 'OUTPUT', 'PREROUTING', 'POSTROUTING', or 'FORWARD'
    :param version: Transport version, must be 'ipv4' or 'ipv6'
    :param rules: List of IPRule objects
    :param _id: Id for this object
    :return:
    """
    obj = IPRing({u'name': name, u'version': version, u'id': _id})
    obj.rules = rules
    return obj


def get_rule(desc=None, ifacein_name=None, ifacein_invert=None, ifaceout_name=None,
             ifaceout_invert=None, ip_protocol_name=None, ip_protocol_invert=None, source_address=None,
             source_mask=None, source_invert=None, dest_address=None, dest_mask=None, dest_invert=None,
             fragment=None, fragment_invert=None, enabled=True, sortId=0, matches=None, jump=None, _id=0):
    """
    Return an IPRule object
    :param matches: List of IPMatch objects
    :param jump: Single IPJump object
    :return:
    """
    obj = IPRule({u'desc': desc,
                  u'ifacein_name': ifacein_name,
                  u'ifacein_invert': ifacein_invert,
                  u'ifaceout_name': ifaceout_name,
                  u'ifaceout_invert': ifaceout_invert,
                  u'ip_protocol_name': ip_protocol_name,
                  u'ip_protocol_invert': ip_protocol_invert,
                  u'source_address': source_address,
                  u'source_mask': source_mask,
                  u'source_invert': source_invert,
                  u'dest_address': dest_address,
                  u'dest_mask': dest_mask,
                  u'dest_invert': dest_invert,
                  u'fragment': fragment,
                  u'fragment_invert': fragment_invert,
                  u'enabled': enabled,
                  u'sortId': sortId,
                  u'id': _id})
    obj.matches = matches
    obj.jump = jump
    return obj


def get_match(name, options, _id=0):
    """

    :param name: Match name
    :param options: List of IPMatchOptions objects
    :param _id: Id for this object
    :return: IPMatch object
    """
    obj = IPMatch({u'name': name, u'id': _id})
    obj.options = options
    return obj


def get_match_option(option, value, invert=False, sort_id=0, _id=0):
    """

    :param option: Option name
    :param value: Option value
    :param invert: Invert option meaning
    :param sort_id: sorting value for this object
    :param _id: Id for this object
    :return: IPMatchOptions object
    """
    return IPMatchOptions({u'option': option, u'value': value, u'invert': invert, u'sortId': sort_id, u'id': _id})


def get_jump(target='ACCEPT', params=None, _id=0):
    """
    Return an IPJump object
    :param target: Jump target value.
    :param params: IPJumpOptions object
    :param _id: Id for this object
    :return: IPJump object
    """
    obj = IPJump({u'id': _id, u'target': target})
    obj.params = params
    return obj


def get_jump_option(name, value=None, _id=0):
    """
    Return a IPJumpOptions object
    :param name: Jump option name
    :param value: Jump option value
    :param _id: Id for this object
    :return: IPJumpOptions object
    """
    return IPJumpOptions({u'name': name, u'value': value, u'id': _id})



#
# Authors: Ma He <mahe.itsec@gmail.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import hashlib
import json
import logging
import os
import socket
import sys
import time

try:
    from urllib import parse
except ImportError:
    # < 3.0
    import urlparse as parse

import dns.resolver
import requests
from silentdune_client.utils.misc import is_valid_ipv4_address, is_valid_ipv6_address
from silentdune_client.builders import iptables as ipt
from silentdune_client.models.remote_config import RemoteConfig
from silentdune_client.modules import BaseModule, QueueTask
from silentdune_client.modules.firewall.manager import SilentDuneClientFirewallModule, TASK_FIREWALL_INSERT_RULES, \
    TASK_FIREWALL_DELETE_SLOT, TASK_FIREWALL_ALLOW_ALL_DNS_ACCESS, TASK_FIREWALL_DISABLE_ALL_DNS_ACCESS
from silentdune_client.modules.firewall.manager.iptables_utils import create_iptables_egress_ingress_rule, \
    create_iptables_remote_config, create_iptables_egress_rule_dest, create_iptables_ingress_rule_source
from silentdune_client.modules.firewall.manager.slots import Slots

_logger = logging.getLogger('sd-client')

module_list = {
    'Silent Dune Remote Config': {
        'module': 'SilentDuneRemoteConfigModule',
    },
}


class SilentDuneRemoteConfigModule(BaseModule):
    """Silent Dune Remote Config Module"""

    _t_check_remote_config = 0  # Timed event property for debug message loop.
    _t_reload_task = 0  # Timed event property for reloading firewall rules.

    _uri = None
    _remote_config_url = None
    _remote_config_dns_name = None
    _remote_config_check_interval = 0

    _remote_config_hash = None

    _slot_config_access = Slots.rc_access
    _slot_config_apply = Slots.rc_apply

    _startup = True

    _retry_count = 0

    priority = 90  # Low priority startup

    def __init__(self):

        self._arg_name = 'remoteconfig'  # Set argparser name
        self._config_section_name = 'remote_config'  # Set configuration file section name

        # Enable multi-threading
        self.wants_processing_thread = True

        self._enabled = False
        self._remote_config_check_interval = 4 * 3600  # 4 hours

    def add_installer_arguments(self, parser):
        """
        Virtual Override
        Add our module's argparser arguments
        """

        # Create a argument group for our module
        group = parser.add_argument_group('remote config module', 'Silent Dune Remote Config module')

        group.add_argument('--remoteconfig-mod-enable', action='store_true',  # noqa
                           help=_('Enable the remote config module'))  # noqa

        group.add_argument(_('--remote-config-url'), help=_('Set remote config url'),  # noqa
                           default=None, type=str)  # noqa
        group.add_argument(_('--remote-config-dns-name'), help=_('Remote config hash value DNS TXT field.'),  # noqa
                           default=None, type=str)  # noqa
        group.add_argument(_('--remote-config-check-interval'), help=_('Set check interval for remote config'),  # noqa
                           default=4 * 3600, type=str)  # noqa

    def validate_arguments(self, args):
        """
        Virtual Override
        Validate command line arguments and save values to our configuration object.
        :param args: An argparse object.
        """
        # See if we have been enabled or not
        if '--remoteconfig-mod-enable' not in sys.argv:
            return True

        if args.remoteconfig_mod_enable:
            self._enabled = True

        # Check for any required arguments here
        try:
            self._uri = parse.urlparse(args.remote_config_url)
        except:
            print('sdc-install: error validating remote config url ({0})'.format(args.remote_config_url))
            return False

        if not self._uri.scheme or self._uri.scheme not in ['file', 'http', 'https', 'ftp']:
            print('sdc-install: invalid protocol for remote config url ({0})'.format(self._uri.scheme))
            return False

        if self._uri.scheme == 'file':
            if not os.path.isfile(self._uri.path):
                print('sdc-install: warning: rule config file not found ({1})'.format(self.get_name(), self._uri.path))
                
        self._remote_config_url = args.remote_config_url
        self._remote_config_dns_name = args.remote_config_dns_name
        self._remote_config_check_interval = args.remote_config_check_interval

        return True

    def validate_config(self, config):
        """
        Virtual Override
        Validate configuration file arguments and save values to our config object.
        :param config: A ConfigParser object.
        """

        # See if we are enabled or not
        try:
            self._enabled = True if config.get(self._config_section_name, 'enabled').lower() == 'yes' else False
        except:
            _logger.debug('{0} configuration section not found in configuration file.'.format(
                self._config_section_name))
            self._enabled = False

        # Only worry about the rest of the configuration items if we are enabled.
        if self._enabled:

            self._remote_config_url = config.get(self._config_section_name, 'remote_config_url').lower()

            try:
                self._uri = parse.urlparse(self._remote_config_url)
            except:
                _logger.error('{0}: error validating remote config url ({1})'.format(self.get_name(),
                                                                                     self._remote_config_url))
                return False

            if not self._uri.scheme or self._uri.scheme not in ['file', 'http', 'https', 'ftp']:
                _logger.error('{0}: invalid protocol for remote config url ({1})'.format(self.get_name(),
                                                                                         self._uri.scheme))
                return None

            if self._uri.scheme == 'file':
                if not os.path.isfile(self._uri.path):
                    _logger.warning('{0}: rule config file not found ({1})'.format(self.get_name(), self._uri.path))

            self._remote_config_url = config.get(self._config_section_name, 'remote_config_url').lower()
            self._remote_config_dns_name = config.get(self._config_section_name, 'remote_config_dns_name').lower()
            self._remote_config_check_interval = int(
                config.get(self._config_section_name, 'remote_config_check_interval'))

        return True

    def prepare_config(self, config):
        """
        Virtual Override
        Return the configuration file structure. Any new configuration items should be added here.
        Note: The order should be reverse of the expected order in the configuration file.
        """

        config.set(self._config_section_name, 'enabled', 'yes' if self._enabled else 'no')
        config.set(self._config_section_name, 'remote_config_url',
                   self._remote_config_url if self._remote_config_url else None)
        config.set(self._config_section_name, 'remote_config_dns_name',
                   self._remote_config_dns_name if self._remote_config_dns_name else '')
        config.set(self._config_section_name, 'remote_config_check_interval',
                   self._remote_config_check_interval if self._remote_config_check_interval else 4 * 3600)

        config.set_comment(self._config_section_name, self._config_section_name,
                           _('; Silent Dune remote config Module Configuration\n'))  # noqa
        config.set_comment(self._config_section_name, 'remote_config_url',
                           _('; Remote config url.\n'))  # noqa
        config.set_comment(self._config_section_name, 'remote_config_dns_name',
                           _('; DNS TXT field to retrieve remote config hash check value.\n'))  # noqa
        config.set_comment(self._config_section_name, 'remote_config_check_interval',
                           _('; Interval in seconds to re-check remote config.\n'))  # noqa

        return True

    def service_startup(self):
        _logger.debug('{0}: module startup called'.format(self.get_name()))

        if not self.validate_config(self.config):
            _logger.error('{0}: module configuration validation failed.'.format(self.get_name()))
            return False

        if not self._enabled:
            return True

        _logger.debug(
            '{0} config: {1}: {2}: {3}'.format(self.get_name(), self._remote_config_url,
                                               self._remote_config_dns_name,
                                               self._remote_config_check_interval))

        return True

    def service_shutdown(self):
        _logger.debug('{0}: module shutdown called'.format(self.get_name()))

    def process_loop(self):

        if self._startup:

            if self._uri.scheme == 'file':
                self.apply_remote_config()
                self._startup = False
                return

            self._retry_count = 0

            # Retry lookup up to 10 times
            while self._retry_count < 10:
                self._retry_count += 1

                # Try resolving the config file host and setup rules
                try:
                    socket.getaddrinfo(self._uri.hostname, 80, 0, 0, socket.IPPROTO_TCP)

                    self.allow_remote_config_access()
                    self.apply_remote_config()

                    self._startup = False

                    return

                except:
                    pass

                if self._retry_count >= 10:
                    _logger.error('{0}: unable to resolve host ({1})'.format(self.get_name(), self._uri.hostname))
                    _logger.error('{0}: unable to download remote config file from host'.format(self.get_name()))
                    self._startup = False
                    return

                time.sleep(1)

        if self.timed_event('_t_check_remote_config', self._remote_config_check_interval):
            self._startup = True

    def allow_remote_config_access(self):

        rules = list()

        if self._uri.scheme == 'http':
            rules.append(self.get_http_rule_by_hostname())
        elif self._uri.scheme == 'https':
            rules.append(self.get_https_rule_by_hostname())
        elif self._uri.scheme == 'ftp':
            rules.append(self.get_ftp_rule_by_hostname())

        # Notify the firewall module to reload the rules.
        task = QueueTask(TASK_FIREWALL_INSERT_RULES,
                         src_module=self.get_name(),
                         dest_module=SilentDuneClientFirewallModule().get_name(),
                         data=rules)
        self.send_parent_task(task)

    def get_http_rule_by_hostname(self):

        rules = list()

        port = self._uri.port
        if not port:
            port = 80

        ipaddrs = self.get_ip_from_hostname(self._uri.hostname, None)

        if not ipaddrs:
            _logger.error('{0}: no ip addresses found from lookup, this is unexpected.'.format(self.get_name()))
            return None

        for ipaddr in ipaddrs:
            _logger.debug(
                '{0}: adding rules for: hostname: {1}, ip addr: {2}, port: {3}'.format(
                    self.get_name(), self._uri.hostname, ipaddr, port))
            rules.append(create_iptables_egress_ingress_rule(ipaddr, port, 'tcp', self._slot_config_access,
                                                             transport=ipt.TRANSPORT_AUTO))

        return rules

    def get_https_rule_by_hostname(self):

        rules = list()

        port = self._uri.port
        if not port:
            port = 443

        ipaddrs = self.get_ip_from_hostname(self._uri.hostname, None)

        if not ipaddrs:
            _logger.error('{0}: no ip addresses found from lookup, this is unexpected.'.format(self.get_name()))
            return None

        for ipaddr in ipaddrs:
            _logger.debug(
                '{0}: adding rules for: hostname: {1}, ip addr: {2}, port: {3}'.format(
                    self.get_name(), self._uri.hostname, ipaddr, port))
            rules.append(create_iptables_egress_ingress_rule(ipaddr, port, 'tcp', self._slot_config_access,
                                                             transport=ipt.TRANSPORT_AUTO))

        return rules

    def get_ftp_rule_by_hostname(self):

        rules = list()

        port = self._uri.port
        if not port:
            port = 21

        ipaddrs = self.get_ip_from_hostname(self._uri.hostname, None)

        if not ipaddrs:
            _logger.error('{0}: no ip addresses found from lookup, this is unexpected.'.format(self.get_name()))
            return None

        for ipaddr in ipaddrs:
            _logger.debug(
                '{0}: adding rules for: hostname: {1}, ip addr: {2}, ftp ports: {3}, 20, etc'.format(
                    self.get_name(), self._uri.hostname, ipaddr, port))
            # FTP control
            rules.append(create_iptables_egress_ingress_rule(ipaddr, port, u'tcp', self._slot_config_access,
                                                             transport=ipt.TRANSPORT_AUTO))
            # FTP data transfer
            rules.append(
                create_iptables_egress_rule_dest(ipaddr, 20, u'tcp', self._slot_config_access, u'ESTABLISHED',
                                                 transport=ipt.TRANSPORT_AUTO))
            rules.append(create_iptables_ingress_rule_source(ipaddr, 20, u'tcp', self._slot_config_access,
                                                             u'ESTABLISHED,RELATED', transport=ipt.TRANSPORT_AUTO))
            rules.append(create_iptables_egress_rule_dest(ipaddr, None, u'tcp', self._slot_config_access,
                                                          u'ESTABLISHED,RELATED', transport=ipt.TRANSPORT_AUTO))
            rules.append(
                create_iptables_ingress_rule_source(ipaddr, None, u'tcp', self._slot_config_access, u'ESTABLISHED',
                                                    transport=ipt.TRANSPORT_AUTO))

        return rules

    def get_ip_from_hostname(self, hostname, port):

        ipaddrs = list()

        if not hostname:
            _logger.debug('{0}: no hostname specified, unable to do lookup'.format(self.get_name()))
            return None

        if port and not (1 < port <= 65536):
            _logger.debug('{0}: port value out of range, unable to do lookup'.format(self.get_name()))
            return None

        try:
            ais = socket.getaddrinfo(hostname, port, 0, 0, socket.IPPROTO_TCP)

            for result in ais:
                ipaddr = result[-1][0]
                ipaddrs.append(ipaddr)

        except:
            _logger.error('{0} unable to resolve hostname ({0})'.format(self.get_name(), hostname))
            return None

        return ipaddrs

    def apply_remote_config(self):

        rules = list()

        try:

            if self._uri.scheme == 'file':
                with open(self._uri.path) as handle:
                    rq = handle.read().decode('utf-8')
            else:

                rq = requests.get(self._remote_config_url)

                if rq.status_code != 200:
                    _logger.error('{0}: unable to retrieve remote configuration'.format(self.get_name()))
                    return
        except:
            _logger.error('{0}: unable to retrieve remote config rules'.format(self.get_name()))
            return

        try:
            if self._uri.scheme == 'file':
                rc = RemoteConfig(json.loads(rq))
                rule_text = rq
            else:
                rc = RemoteConfig(rq.json())
                rule_text = rq.text

        except:
            _logger.error('{0}: parsing remote configuration file failed'.format(self.get_name()))
            return

        # See if we should check the hash value against a DNS TXT lookup hash value
        if self._remote_config_dns_name:

            if not rc.hash_method or rc.hash_method not in ['sha1', 'sha256', 'sha512']:
                _logger.error('{0}: unsupported hashing algorithm in remote config ({1})'.format(
                    self.get_name(), rc.hash_method))
                return

            if rc.hash_method == 'sha1':
                hash_value = hashlib.sha1(rule_text).hexdigest()
            elif rc.hash_method == 'sha256':
                hash_value = hashlib.sha256(rule_text).hexdigest()
            elif rc.hash_method == 'sha512':
                hash_value = hashlib.sha512(rule_text).hexdigest()

            if self.get_remote_dns_hash() != hash_value:
                _logger.error('{0}: remote DNS TXT hash value does not match calculated hash'.format(self.get_name()))
                return

        _logger.debug('{0}: deleting rules for slot {1}'.format(self.get_name(), self._slot_config_apply))
        task = QueueTask(TASK_FIREWALL_DELETE_SLOT,
                         src_module=self.get_name(),
                         dest_module=SilentDuneClientFirewallModule().get_name(),
                         data=self._slot_config_apply)
        self.send_parent_task(task)

        for rule in rc.rules:

            try:

                # See if we need to try and resolve the host name
                if not is_valid_ipv6_address(rule.host) and not is_valid_ipv6_address(rule.host):

                    addrs = self.get_ip_from_hostname(rule.host, rule.port)

                    if not addrs:
                        _logger.error('{0}: error: unable to resolve host {1}'.format(rule.host))
                        return

                    for addr in addrs:
                        _logger.debug('{0}: adding rules slot: {1}, ip address: {2}, port: {3}, protocol: {4}'.format(
                            self.get_name(), self._slot_config_apply, addr, rule.port, rule.protocol))
                        rules.append(create_iptables_remote_config(addr, rule.mask, rule.port, rule.protocol,
                                                                   self._slot_config_apply, rule.uid, rule.gid,
                                                                   transport=ipt.TRANSPORT_AUTO))

                else:
                    _logger.debug('{0}: adding rules slot: {1}, ip address: {2}, port: {3}, protocol: {4}'.format(
                        self.get_name(), self._slot_config_apply, rule.host, rule.port, rule.protocol))
                    rules.append(create_iptables_remote_config(rule.host, rule.mask, rule.port, rule.protocol,
                                                               self._slot_config_apply, rule.uid, rule.gid,
                                                               transport=ipt.TRANSPORT_AUTO))

            except:
                _logger.error('{0}: parsing remote config rule failed.'.format(self.get_name()))

        if rules:
            task = QueueTask(TASK_FIREWALL_INSERT_RULES,
                             src_module=self.get_name(),
                             dest_module=SilentDuneClientFirewallModule().get_name(),
                             data=rules)
            self.send_parent_task(task)

    def get_remote_dns_hash(self):
        """
        Retrieve the remote config hash value stored in the given DNS TXT field name
        :return:
        """
        query = self._remote_config_dns_name

        if not query:
            return None

        try:
            answers = dns.resolver.query(query, 'TXT')

            for rdata in answers:
                for txt_string in rdata.strings:
                    return txt_string

        except:
            _logger.error('{0}: remote config DNS hash retrieval failed ({1})'.format(self.get_name(), query))

        return None



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
import random
import shlex
import subprocess
import string
import sys

from silentdune_client.utils.misc import \
    which, determine_config_root, get_active_firewall, get_init_system, is_service_running

_logger = logging.getLogger('sd-client')


class NodeInformation:

    # External programs
    ps = None
    pgrep = None
    sed = None
    kill = None
    tail = None

    # SELinux
    restorecon_exec = None
    chcon_exec = None

    # ipv4 commands
    iptables_exec = None
    iptables_save = None
    iptables_restore = None

    # ipv6 commands
    ip6tables_exec = None
    ip6tables_save = None
    ip6tables_restore = None

    # Configuration Items
    root_user = (os.getuid() == 0)
    config_root = None
    machine_id = None
    pid_file = None
    log_file = None

    # Upstart
    ups_installed = False

    # Systemd
    sysd_installed = False

    # sysvinit
    sysv_installed = False

    # Previous firewall service
    previous_firewall_service = None

    # Firewall_platform - Currently only iptables is supported.
    firewall_platform = 'iptables'

    # OS Platform - IE: linux, osx, ...
    platform = 'linux'

    error = False

    console_debug = False

    def __init__(self, console_debug=False):
        """
        Gather all the information we need about this node.
        """

        self.console_debug = console_debug

        if sys.platform == "linux" or sys.platform == "linux2":
            self.platform = 'linux'
        elif sys.platform == "darwin":
            self.platform = 'osx'  # OS X
        elif sys.platform == "win32":
            self.platform = 'windows'

        # Gather information about this node.
        self.config_root = determine_config_root()
        if not self.config_root:
            if self.console_debug:
                print('Error: Unable to determine configuration root.')
            self.error = True
            return

        # Check to see if our configuration path is in a home directory.
        if os.path.expanduser('~') in self.config_root:
            _logger.warning('Not running as root, setting configuration path to "{0}"'.format(self.config_root))
            _logger.warning('Since we are not running as root, system firewall settings will not be changed.')

        # Change the default path for pid and log file to config_root.
        self.pid_file = os.path.join(self.config_root, 'sdc.pid')
        self.log_file = os.path.join(self.config_root, 'sdc.log')

        if not self._check_for_external_progs():
            self.error = True
        elif not self._init_system_check():
            self.error = True
        elif not self._firewall_check():
            self.error = True
        elif not self._get_machine_id():
            self.error = True

    def _which_wrapper(self, name):
        """
        Look for program
        :param name: Program name
        :return: Program path or None
        """

        p = which(name)
        if not p:
            if self.console_debug:
                # print('Failed to find program path for "{0}"'.format(name))
                pass
            else:
                _logger.warning('failed to find program path for "{0}"'.format(name))

        return p

    def _check_for_external_progs(self):
        """
        Find external programs used by this client.
        """

        # Use _which_wrapper to set an error for any missing programs.
        self.ps = self._which_wrapper('ps')
        self.pgrep = self._which_wrapper('pgrep')
        self.sed = self._which_wrapper('sed')
        self.kill = self._which_wrapper('kill')
        self.tail = self._which_wrapper('tail')

        # SElinux
        self.restorecon_exec = self._which_wrapper('restorecon')
        self.chcon_exec = self._which_wrapper('chcon')

        self.iptables_exec = self._which_wrapper('iptables')
        self.iptables_save = self._which_wrapper('iptables-save')
        self.iptables_restore = self._which_wrapper('iptables-restore')

        self.ip6tables_exec = self._which_wrapper('ip6tables')
        self.ip6tables_save = self._which_wrapper('ip6tables-save')
        self.ip6tables_restore = self._which_wrapper('ip6tables-restore')

        if self.ps is None or self.pgrep is None:
            if self.console_debug:
                print('Error: Unable to determine which services are running on this machine.')
            else:
                _logger.error('Unable to determine which services are running on this machine.')
            return False

        if self.sed is None or self.kill is None or self.tail is None:
            if self.console_debug:
                print('Crit: Missing system critical system executables.')
            else:
                _logger.critical('Missing system critical system executables.')

        if self.iptables_exec is None or self.iptables_save is None or self.iptables_restore is None:
            if self.console_debug:
                print('Unable to find iptables executables.')
            else:
                _logger.debug('Crit: Unable to find iptables executables.')
            return False

        return True

    def _init_system_check(self):
        """
        Determine which init system is running on this system.
        """

        init = get_init_system()

        # If we didn't detect the init system, abort.
        if not init:
            if self.console_debug:
                print('Error: Unable to determine Init version')
            else:
                _logger.debug('Error: Unable to determine Init version')
            return False

        if init == 'upstart':
            self.ups_installed = True

        if init == 'systemd':
            self.sysd_installed = True

        if init == 'sysv':
            self.sysv_installed = True

        return True

    def _firewall_check(self):
        """
        Get the currently running firewall service
        """

        self.previous_firewall_service = get_active_firewall()

        if not self.previous_firewall_service:
            return True

        if not is_service_running(self.previous_firewall_service):
            if self.console_debug:
                print('Info: no firewall service is currently running.')
            else:
                _logger.debug('Info: no firewall service is currently running.')

        return True

    def _get_machine_id(self):
        """
        Find the machine unique identifier or generate one for this machine.
        """

        m_id = None
        f = os.path.join(self.config_root, 'machine-id')

        # See if we have an exiting machine-id file in our config root
        if os.path.exists(f):
            with open(f, 'r') as handle:
                m_id = handle.readline().strip('\n')

        if not m_id:

            # See if we can find a machine-id file on this machine
            for p in ['/etc/machine-id', '/var/lib/dbus/machine-id']:
                if os.path.isfile(p):
                    with open(p) as handle:
                        m_id = handle.readline().strip('\n')

        # If we can't find an existing machine id, make one up.
        if not m_id:
            if self.console_debug:
                print('Info: Existing machine id not found, generating new machine id.')
            else:
                _logger.info('Existing machine id not found, generating new machine id.')
            m_id = self._write_machine_id()

        self.machine_id = m_id

        return True

    def _write_machine_id(self):
        """
        Create a new unique machine id for this node.
        """
        m_id = ''.join(random.choice('abcdef' + string.digits) for _ in range(32))

        if not os.path.isdir(self.config_root):
            os.makedirs(self.config_root, 0o770)

        mfile = os.path.join(self.config_root, 'machine-id')
        with open(mfile, 'w') as h:
            h.write(m_id + '\n')

        os.chmod(mfile, 0o440)

        return m_id

    def _run_service_command(self, cmd, name):
        """
        Helper function for running system service commands.
        """
        # SysV and Upstart default
        prog = which('service')
        args = [prog, name, cmd]  # Note order of 'name' and 'cmd'

        # SystemD
        if self.sysd_installed:
            prog = which('systemctl')
            args = [prog, cmd, name, '2>/dev/null']  # Note order of 'name' and 'cmd'

        p = subprocess.Popen(args, stdout=subprocess.PIPE)
        stdoutdata, stderrdata = p.communicate()
        p.wait()

        if stderrdata or p.returncode != 0:
            return False

        return True

    def start_service(self, name):
        """
        Start a system service.
        :param name: Name of service
        :return: True if successful, otherwise False
        """
        return self._run_service_command('start', name)

    def stop_service(self, name):
        """
        Stop a system service.
        :param name: Name of service
        :return: True if successful, otherwise False
        """
        return self._run_service_command('stop', name)

    def disable_service(self, name):
        """
        Disable a system service.
        :param name: Name of service
        :return: True if successful, otherwise False
        """
        # Sys V
        if self.sysv_installed:
            return self._run_service_command('disable', name)

        # System D mask service
        if self.sysd_installed:
            if not self._run_service_command('disable', name):
                return False
            return self._run_service_command('mask', name)

        if self.ups_installed:
            # Upstart requires an override file to disable the service
            try:
                with open('/etc/init/{0}.override'.format(name), 'w') as h:
                    h.write('manual')
                return True
            except IOError:
                _logger.error('Unable to create upstart service override file.')

        return False

    def enable_service(self, name):
        """
        Enable a system service.
        :param name: Name of service
        :return: True if successful, otherwise False
        """
        # Sys V
        if self.sysv_installed:
            return self._run_service_command('enable', name)

        # SystemD service
        if self.sysd_installed:
            self._run_service_command('unmask', name)
            return self._run_service_command('enable', name)

        # Upstart service
        if self.ups_installed:

            file = '/etc/init/{0}.override'.format(name)
            # Check to see if the override file exists.
            if os.path.exists(file):
                os.remove(file)
            else:

                args = [self.sed, "-i '/manual/d' '/etc/init/{0}'".format(name)]

                p = subprocess.Popen(args, stdout=subprocess.PIPE)
                stdoutdata, stderrdata = p.communicate()
                p.wait()

                if stderrdata or p.returncode != 0:
                    return False

                return True

        return False

    def restorecon(self, path):

        if not path:
            return False

        # Restore SElinux contexts if selinux is available
        if self.restorecon_exec and self.chcon_exec:
            cmd = '{0} -R {1}'.format(self.restorecon_exec, path)

            try:
                p = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE)
                stdoutdata, stderrdata = p.communicate()
                p.wait()

                if stderrdata:
                    return False

                # Set selinux user context for path
                cmd = '{0} -R -u system_u {1}'.format(self.chcon_exec, path)

                with open(os.devnull, 'w') as FNULL:
                    p = subprocess.Popen(shlex.split(cmd), stdout=FNULL, stderr=FNULL)
                    p.communicate()
                    p.wait()

            except:
                return False

        return True



#
# Authors: Robert Abram <robert.abram@entpack.com>,
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
import time

from silentdune_client.utils.node_info import NodeInformation
from silentdune_client.utils.console import ConsoleBase
from silentdune_client.utils.module_loading import import_by_str
from silentdune_client.utils.exceptions import ModuleLoadError

_logger = logging.getLogger('sd-client')

# Parent to Child task queue IDs
TASK_STOP_PROCESSING = 0


class QueueTask(object):
    """
    The QueueTask object is an object to be passed between the parent processing thread
    and the module processing threads.  Allows for back and forth communication along with
    passing tasks between module processing threads.
    """

    _task_id = None  # One of the TASK id value from above.
    _trans_id = None  # ID used by the source to track messages.
    _src_module = None  # Source module name
    _dest_module = None  # Destination module name
    _data = None  # Data for this task.

    def __init__(self, task_id, src_module=None, dest_module=None, data=None, trans_id=None, ):
        self._task_id = task_id
        self._src_module = src_module
        self._dest_module = dest_module
        self._data = data
        self._trans_id = trans_id

    def get_task_id(self):
        return self._task_id

    def get_sender(self):
        return self._src_module

    def get_trans_id(self):
        return self._trans_id

    def get_dest_name(self):
        return self._dest_module

    def get_data(self):
        return self._data


class BaseModule(ConsoleBase):
    """
    This is the Virtual module object every module should inherit from.
    Each property and method are virtual and can be overridden as needed.
    """

    # The name of the module and version.
    # _name = 'UnknownModule'
    _arg_name = 'unknown'  # This is the argparser name for this module
    _config_section_name = 'unknown'  # This is the configuration file section name
    _version = '0.0.1'
    _enabled = True

    # Multi thread processing properties
    wants_processing_thread = False  # Set this to true if your module will use a child processing thread.

    # Parent process management queue.  This is only valid if we are using a processing thread.
    _mqueue = None

    # List of loaded module names, useful to see if another module has been loaded.
    _mlist = None

    # Delay in seconds to wait for a message from the parent.  After timeout has expired, process_loop is called.
    _queue_timeout = 1.0  # Min = 0.01, Max = 10.0.

    # _start_t and _seconds_t can be used to for timed events.
    t_start = time.time()
    t_seconds = 0                   # Number of seconds that have passed since this process started.

    # Configuration File and Node Information objects.
    config = None
    node_info = NodeInformation()

    # Module loading priority
    priority = 100  # 0 highest -> 100 lowest

    """
    Installer Virtual Methods
    """

    def get_name(self):
        """
        :return: module name
        """
        # return self._name
        return type(self).__name__

    def get_version(self):
        """
        :return: module version
        """
        return self._version

    def add_installer_arguments(self, parser):
        pass

    def get_config(self):
        """
        :return: configuration
        """
        return self.config

    def disable_module(self):
        self._enabled = False

    def module_enabled(self):
        return self._enabled

    def validate_arguments(self, args):
        """
        Validate command line arguments and save values to our configuration object.
        :param args: An argparse object.
        :return: True if command line arguments are valid, otherwise False.
        """
        pass

    def validate_config(self, config):
        """
        Validate configuration file arguments and save values to our config object.
        :param config: A ConfigParser object.
        :return: True if configuration file values are valid, otherwise False.
        """
        pass

    def prepare_config(self, config):
        """
        Add the module configuration items that need to be saved to the configuration file.
        :param config: A ClientConfiguration object.
        :return: True if configuration file values were prepared correctly, otherwise False.
        """
        pass

    def set_config(self, config):
        """
        !!! Please Do not override this method
        """
        self.config = config
        if self.validate_config(config) is False:
            self.disable_module()

    def pre_install(self):
        """
        Called by the installer before the formal install process starts.
        :param installer: The Installer object.
        :return: True if successful, otherwise False.
        """
        pass

    def install_module(self):
        """
        Called by the installer during the formal install process.
        :param installer: The Installer object.
        :return: True if successful, otherwise False.
        """
        pass

    def post_install(self):
        """
        Called by the installer after the formal install process has completed.
        :param installer: The Installer object.
        :return: True if successful, otherwise False.
        """
        pass

    def uninstall_module(self):
        """
        Called by the installer during an uninstall process.
        :param installer: The Installer object.
        :return: True if successful, otherwise False.
        """
        pass

    """
    Service Daemon Virtual Methods
    """

    def service_startup(self):
        """
        Called by the service daemon during service start or reload.
        :return: True if successful, otherwise False.
        """
        pass

    def service_shutdown(self):
        """
        Called by the service daemon during service stop.
        :return: True if successful, otherwise False.
        """
        pass

    def process_task(self, task):
        """
        Process a QueueTask object and do something. Called by the process_handler method
        when there is a QueueTask object sent by the parent process.
        :param task:
        :return:
        """
        pass

    def process_loop(self):
        """
        This is called during after each idle timeout period has expired in process_handler.
        This method should do some work if needed and then return to the process_handler.
        Do not set a long term loop in this method. Doing so will break the parent processing.
        :return:
        """
        pass

    def timed_event(self, t_property, seconds):
        """
        Check to see if the number of seconds has passed based on the object property value.
        :param property: Property of the object that is keeping track of the last event time.
        :param seconds: Number of seconds that need to pass before the next event time.
        :return: True if event time is now, False if not.
        """
        # Get the property value
        t_property_val = getattr(self, t_property)

        # Check to see if we have reached the next event time.
        if self.t_seconds > t_property_val and self.t_seconds % seconds == 0.0:
            setattr(self, t_property, self.t_seconds)
            return True

        return False

    def process_handler(self, queue, mqueue, mlist):
        """
        !!! Please do not override this method, override the process_loop method !!!
        Called during the service loop.
        :param queue: Multiprocessing queue object.
        """
        _logger.debug('{0} processing thread started.'.format(self.get_name()))

        # Save parent's manager queue.
        self._mqueue = mqueue

        # Save list of all loaded module names.
        self._mlist = mlist

        while True:

            try:
                task = queue.get(timeout=self._queue_timeout)  # Wait while looking for a QueueTask object.
            except:
                self.t_seconds = int(time.time() - self.t_start)  # Set the number of seconds since process start.
                self.process_loop()  # Call the processing loop for module idle processing.
                continue

            try:
                # _logger.debug('{0}: received task id: {1}.'.format(self.get_name(), task.get_task_id()))
                if task.get_task_id() == TASK_STOP_PROCESSING:
                    _logger.debug('{0}: received stop signal, ending process handler.'.format(self.get_name()))
                    self.service_shutdown()
                    break

                # Process task.
                self.process_task(task)

            except:
                raise
                # _logger.debug('{0}: received bad task object, discarding.'.format(self.get_name()))

        _logger.debug('{0}: Module processing thread closed cleanly.'.format(self.get_name()))

    def send_parent_task(self, qtask):
        """
        Send a QueueTask object to the parent process.
        :param qtask: QueueTask object
        :return:
        """
        if self._mqueue:
            self._mqueue.put(qtask)
        else:
            _logger.error('{0}: Task submitted to parent, but queue not ready.'.format(self.get_name()))

    def __lt__(self, other):
        """Sort modules by priority"""
        return self.priority < other.priority


def __load_modules__(base_path=None, module_path='silentdune_client/modules'):
    """
    Search for modules to load.  Modules must reside under the modules directory and
    have a "module_list" dict defined in the __init__.py file. Each entry in the
    "module_list" must list a Class that subclasses BaseModule.
    """

    module_list = list()

    # Loop through the directories looking for modules to import.
    for root, dirs, files in os.walk(os.path.join(base_path, module_path), topdown=True):
        # Skip our directory.
        if root == '.':
            continue

        # Look only at __init__.py files.
        for name in files:
            if name == '__init__.py':

                # Remove base_path and convert to dotted path.
                mp = root.replace(base_path + '/', '').replace('./', '').replace('/', '.')

                # Attempt to import 'module_list' from __init__.py file.
                try:
                    ml = import_by_str(mp + '.module_list')

                # If we get an Exception check to see if the python module loaded but there was no
                # client module definition found, otherwise just reraise the last Exception for debugging.
                except ModuleLoadError:
                    # Looks like a clean import error. IE: __init__.py is not a real module.
                    continue
                except:
                    # Found a module to load, but it threw an Exception. Just pass the Exception up.
                    raise

                for mname, mdict in ml.items():
                    # _logger.debug('Found module definition "{0}" in path {1}'.format(mname, mp))
                    for key, name in mdict.items():

                        if key == 'module':
                            tpath = mp + '.' + name
                            try:
                                mod = import_by_str(tpath)
                                module_list.append(mod())
                                _logger.debug('Adding "{0}" module ({1}).'.format(mname, tpath))
                            except ImportError:
                                _logger.error('Adding "{0}" module failed. ({1}).'.format(mname, tpath))
                                pass

    return module_list



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from importlib import import_module

from silentdune_client.utils.exceptions import ModuleLoadError


def import_by_str(mod):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """

    # If these throw Exceptions, just let them bubble to the parent.
    mpath, cname = mod.rsplit('.', 1)
    module = import_module(mpath)

    try:
        return getattr(module, cname)
    except AttributeError as e:
        msg = str(e)

        # If this is a "'module' object has no attribute 'module_list'" message raise an ModuleLoadError,
        # otherwise reraise the AttributeError Exception so we can get the cause and stacktrace of the
        # Exception.
        if "no attribute 'module_list'" in msg:
            raise ModuleLoadError('Not a loadable module')
        else:
            raise




#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ast
import hashlib
import logging
import sys

from silentdune_client.modules import BaseModule, QueueTask

from silentdune_client.utils.module_loading import import_by_str
from silentdune_client.modules.firewall.manager import SilentDuneClientFirewallModule, \
    TASK_FIREWALL_INSERT_RULES, TASK_FIREWALL_DELETE_RULES

_logger = logging.getLogger('sd-client')

module_list = {
    'Silent Dune Auto Services': {
        'module': 'SilentDuneAutoServicesModule',
    },
}

AUTO_SERVICES_LIST = (
    "silentdune_client.modules.automation.auto_services.docker.DockerService",
)


class SilentDuneAutoServicesModule(BaseModule):

    _startup = True

    _enabled = False

    _mss_slots = {}  # Storage for machine_subset hashes and slot ids.  For detection of rule changes.

    def __init__(self):

        self._arg_name = 'autoservices'  # Set argparser name
        self._config_section_name = 'auto_services'  # Set configuration file section name

        # Enable multi-threading
        self.wants_processing_thread = True

    def add_installer_arguments(self, parser):
        """
        Virtual Override
        Add our module's argparser arguments
        """

        # Create a argument group for our module
        group = parser.add_argument_group('auto services module', 'Silent Dune Auto Services module')

        group.add_argument('--services-mod-enable', action='store_true',  # noqa
                           help=_('Enable the auto services module'))  # noqa

        # TODO: Add in --services-disable-service as a multi value argparse argument

    def validate_arguments(self, args):
        """
        Virtual Override
        Validate command line arguments and save values to our configuration object.
        :param args: An argparse object.
        """
        # See if we have been enabled or not
        if '--services-mod-enable' not in sys.argv:
            return True

        if args.services_mod_enable:
            self._enabled = True

        return True

    def validate_config(self, config):
        """
        Virtual Override
        Validate configuration file arguments and save values to our config object.
        :param config: A ConfigParser object.
        """

        # See if we are enabled or not
        try:
            self._enabled = True if config.get(self._config_section_name, 'enabled').lower() == 'yes' else False
        except:
            _logger.debug('{0} configuration section not found in configuration file.'.format(
                self._config_section_name))
            self._enabled = False

    def prepare_config(self, config):
        """
        Virtual Override
        Return the configuration file structure. Any new configuration items should be added here.
        Note: The order should be reverse of the expected order in the configuration file.
        """

        config.set(self._config_section_name, 'enabled', 'yes' if self._enabled else 'no')

        config.set_comment(self._config_section_name, self._config_section_name,
                           _('; Silent Dune Auto Services Module Configuration\n'))  # noqa

    def service_startup(self):
        _logger.debug('{0}: module startup called'.format(self.get_name()))

        if not self.validate_config(self.config):
            _logger.error('{0}: module configuration validation failed.'.format(self.get_name()))
            return False

        if not self._enabled:
            return True

        return True

    def service_shutdown(self):
        _logger.debug('{0}: module shutdown called'.format(self.get_name()))

    def process_loop(self):

        if self._startup:
            self.find_active_services()

    def find_active_services(self):
        """
        Loop through the services list and call discover() for each class found.
        """
        for name in AUTO_SERVICES_LIST:
            rule_count = self.check_service(name)

    def check_service(self, name):
        """
        Check the service for rules and add them to the firewall.
        :param name: Service discovery module name
        """
        module_name, class_name = name.rsplit('.', 1)

        _logger.debug('{0}: Loading service object {1}'.format(self.get_name(), class_name))

        module = import_by_str(name)
        cls = module(config=self.config)
        disabled = getattr(self, cls.get_config_property_name())
        if type(disabled) is str:  # Python 2.7 returns string type from getattr(), Python 3.4 returns bool.
            disabled = ast.literal_eval(disabled)

        # _logger.debug('Property: {0}: Value: {1}'.format(cls.get_config_property_name(), disabled))
        # See if this discovery service has been disabled. Name value must match one of our property names.
        if disabled:
            _logger.debug('{0}: {1} service disabled by config.'.format(self.get_name(), class_name))
            return 0

        rules, slot = cls.discover()

        if rules:

            # See if we already have saved rules for this slot id
            if slot in self._mss_slots:
                if self.rules_have_changed(self._mss_slots[slot], rules):

                    _logger.debug('{0}: {1}: Rules have changed, notifying firewall manager.'.format(
                        self.get_name(), class_name))

                    # Notify the firewall module to delete the old rules.
                    task = QueueTask(TASK_FIREWALL_DELETE_RULES,
                                     src_module=self.get_name(),
                                     dest_module=SilentDuneClientFirewallModule().get_name(),
                                     data=self._mss_slots[slot])
                    self.send_parent_task(task)
                else:
                    return 0

            # Save rules so we can check against them next time.
            self._mss_slots[slot] = rules

            # Notify the firewall module to reload the rules.
            task = QueueTask(TASK_FIREWALL_INSERT_RULES,
                             src_module=self.get_name(),
                             dest_module=SilentDuneClientFirewallModule().get_name(),
                             data=rules)
            self.send_parent_task(task)
        else:
            _logger.info('{0}: {1}: service did not return any rules.'.format(
                self.get_name(), class_name))

            return 0

        return len(rules)

    def rules_have_changed(self, old, new):
        """
        Compare rules and see if they are different.
        :param old: List of machine_subset objects
        :param new: List of machine_subset objects
        :return: True if old and new rules are different
        """

        old_rules = ''
        new_rules = ''

        for mss in old:
            old_rules += mss.to_json()

        for mss in new:
            new_rules += mss.to_json()

        return hashlib.sha1(old_rules).hexdigest() != hashlib.sha1(new_rules).hexdigest()




# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from enum import IntEnum


class Slots(IntEnum):
    """ Slot numbers for known rule sets """
    loopback = 0
    admin = 110
    silentdune_server = 120
    dns = 130
    ntp = 140
    updates = 150
    identity = 160
    dhcp = 170
    icmp = 180
    network_iso = 500  # Network isolation rules
    rc_access = 600  # accessing remote config
    rc_apply = 610  # applying remote config
    logging = 9800
    reject = 9900



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


import logging
import os
from collections import OrderedDict

from silentdune_client.utils.misc import determine_config_root, get_active_firewall, rmdir

try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser  # ver. < 3.0

_logger = logging.getLogger('sd-client')

# Global debug variable
debug = False


class ClientConfiguration(object):
    _config_root = None
    _config_file = None  # Full path and file name for the configuration file.
    _config_section = 'settings'  # This is the default configuration file section name, do not change.
    _config = OrderedDict()

    def __init__(self, config_file=None):
        """
        Initialize the configuration object with the most reasonable values for the default settings.
        :param config_file: Use alternate configuration file.
        """

        self._config_root = determine_config_root()

        # Set the configuration file path and name.
        if config_file:
            self._config_file = config_file
            self._config_root = os.path.split(config_file)[0]
        else:
            self._config_file = os.path.join(self._config_root, 'sdc.conf')

        # Set the default file paths based on if we are running as root or not.
        if os.getuid() == 0:
            self.set(self._config_section, 'pidfile', '/var/run/silentdune/sdc.pid')
            self.set(self._config_section, 'logfile', '/var/log/silentdune.log')
        else:
            self.set(self._config_section, 'pidfile', os.path.join(self._config_root, 'sdc.pid'))
            self.set(self._config_section, 'logfile', os.path.join(self._config_root, 'silentdune.log'))

        # self.set(self._config_section, 'user', 'silentdune')
        # self.set(self._config_section, 'group', 'silentdune')

        # # Set the previous firewall service to the currently running firewall.
        # self.set(self._config_section, 'previous_firewall_service', get_active_firewall())

        # Set the section heading comments
        self.set_comment(self._config_section, 'settings',
                         _('; Silent Dune Client Configuration File\n'  # noqa
                           '; This file was automatically generated by the installer.\n'
                           '; Run "sdc-install --help" to configure this file.\n'))

        # Set the section item comments
        self.set_comment(self._config_section, 'pidfile', _('; The path and file name for the PID file.\n'))  # noqa
        # self.set_comment(self._config_section, 'user', _('; The local user the service should run as.\n'  # noqa
        #                                                  '; Default: silentdune.\n'))
        # self.set_comment(self._config_section, 'group', _('; The local group the service should run as.\n'  # noqa
        #                                                   '; Default: silentdune.\n'))
        # self.set_comment(self._config_section, 'previous_firewall_service',
        #                  _('; The previous firewall service. Warning: Changing this value may\n'  # noqa
        #                    '; compromise security on this system if the Silent Dune client is\n'
        #                    '; uninstalled.\n'))

    def delete(self, section, key):
        """
        Delete a configuration key from the section specified.
        :param section: Configuration section name
        :param key:
        :return:
        """

        if not section or section not in self._config:
            msg = 'Invalid configuration section name ({0}).'.format(section)
            raise ValueError(msg)

        if not key or key not in self._config[section]:
            msg = 'Invalid configuration key name ({0}).'.format(key)
            raise ValueError(msg)

        del self._config[section][key]

        return True

    def set(self, section, key, val):
        """
        Set a configuration value in the section specified.
        :param section: Configuration section name
        :param key: Value key name
        :param val: Value to set
        :return:
        """
        if not section:
            msg = 'Invalid configuration section name ({0}).'.format(section)
            raise ValueError(msg)

        if section not in self._config:
            self._config[section] = OrderedDict()

        self._config[section][key] = val

    def get(self, section, key):
        """
        Retrieve a configuration value.
        :param section: Configuration section name
        :param key: Value key name
        :return Value or None:
        """
        if not section or section not in self._config:
            msg = 'Invalid configuration section name ({0}).'.format(section)
            raise ValueError(msg)

        if not key or key not in self._config[section]:
            msg = 'Invalid configuration key name ({0}).'.format(key)
            raise ValueError(msg)

        return self._config[section][key]

    def set_comment(self, section, key, comment):
        """
        Set the comments for a given key in the section specified.
        """

        if not section:
            msg = 'Invalid configuration section name ({0}).'.format(section)
            raise ValueError(msg)

        if 'comments' not in self._config[section]:
            self._config[section]['comments'] = OrderedDict()

        self._config[section]['comments'][key] = comment

    def validate_config_file(self):
        """
        Check to see if the configuration file exists, is readable and valid.
        """

        if not os.path.exists(self._config_file):
            _logger.debug('Configuration file ({0}) does not exist.'.format(self._config_file))
            return None

        config = ConfigParser(dict_type=OrderedDict)

        try:
            config.read(self._config_file)
        except ConfigParser.Error:
            _logger.debug('Configuration file is invalid.')
            return None

        return config

    def read_config(self):
        """
        Read the configuration file.
        """
        # Return the saved configuration loaded by validate_config_file().
        return self.validate_config_file()

    def write_config(self):
        """
        Write out the configuration to a file, we do NOT use ConfigParser to write out the config.
        :param config: Must be a ConfigParser object.
        """
        try:

            with open(self._config_file, 'w') as handle:

                # Loop through each section
                for name, section in self._config.items():
                    _logger.debug('Config: writing section: {0}'.format(name))

                    # Write out the section name
                    handle.write('[' + name + ']\n')

                    # Write any section header comments
                    if name in section['comments']:
                        comment = section['comments'][name]
                        handle.write(comment.rstrip() + '\n\n')

                    # Write each section item out
                    for (key, value) in section.items():

                        if key == 'comments':
                            continue

                        _logger.debug('Config: Key {0}={1}'.format(key, value))

                        # Write item comments if found.
                        if key in section['comments']:
                            comment = section['comments'][key]
                            handle.write(comment.rstrip() + '\n')

                        handle.write(key + '=' + str(value) + '\n\n')

            os.chmod(self._config_file, 0o640)
        except IOError:
            return False

        return True

    def create_directories(self):
        """
        Make sure directories used by the client exist and are set to the correct permissions.
        :return:
        """

        pidfile = None

        try:
            # user = self.get(self._config_section, 'user')
            # group = self.get(self._config_section, 'group')
            logfile = self.get(self._config_section, 'logfile')
        except ValueError:
            _logger.error('Unable to retrieve configuration values.')
            return False

        # The pidfile configuration item might not exist.
        try:
            pidfile = self.get(self._config_section, 'pidfile')
        except ValueError:
            pass

        # try:
        #     # Get information about the daemon process user and group.
        #     userinfo = pwd.getpwnam(user)
        #     groupinfo = grp.getgrnam(group)
        # except:
        #     _logger.error('Unable to get information about daemon process user and group.')
        #     return False

        # Create configuration root directory.
        try:
            configpath = self._config_root
            _logger.debug('Creating Configuration path ({0})'.format(configpath))
            if not os.path.exists(configpath):
                os.makedirs(configpath, 0o770)
            # os.chown(configpath, userinfo.pw_uid, groupinfo.gr_gid)
        except:
            _logger.error('Unable to create PID file path.')
            return False

        # If needed, create PID file path.
        try:
            if pidfile:
                pidpath = os.path.split(pidfile)[0]
                _logger.debug('Creating PID path ({0})'.format(pidpath))
                if not os.path.exists(pidpath):
                    os.makedirs(pidpath, 0o770)
                # os.chown(pidpath, userinfo.pw_uid, groupinfo.gr_gid)
        except:
            _logger.error('Unable to create PID file path.')
            return False

        # Create Log file path.
        try:
            logpath = os.path.split(logfile)[0]
            _logger.debug('Creating Log file path ({0})'.format(logpath))
            if not os.path.exists(logpath):
                os.makedirs(logpath, 0o770)
            # os.chown(logpath, userinfo.pw_uid, groupinfo.gr_gid)
        except:
            _logger.error('Unable to create LOG file path.')
            return False

        return True

    def delete_directories(self):
        """
        Remove directories used by the daemon process
        :return:
        """

        result = True
        pidfile = None

        try:
            logfile = self.get(self._config_section, 'logfile')
        except ValueError:
            _logger.error('Unable to retrieve directory configuration values.')
            return False

        # The pidfile configuration item might not exist.
        try:
            pidfile = self.get(self._config_section, 'pidfile')
        except ValueError:
            pass

        # Remove log file directory
        logpath = os.path.split(logfile)[0]
        if os.path.exists(logpath) and os.path.realpath(logpath) != '/':
            try:
                rmdir(logpath)
            except:
                _logger.error('Failed to delete logfile directory ({0}).'.format(logpath))
                result = False

        # Remove PID file directory
        if pidfile:
            pidpath = os.path.split(pidfile)[0]
            if os.path.exists(pidpath) and os.path.realpath(pidpath) != '/':
                try:
                    rmdir(pidpath)
                except:
                    _logger.error('Failed to delete PID file directory ({0}).'.format(pidpath))
                    result = False

        # Remove Configuration path
        if os.path.exists(self._config_root):
            try:
                rmdir(self._config_root)
            except:
                _logger.error('Failed to delete config directory ({0}).'.format(logpath))
                result = False

        return result



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from silentdune_client.modules import BaseModule, QueueTask
from silentdune_client.modules.firewall.manager import SilentDuneClientFirewallModule, TASK_FIREWALL_RELOAD_RULES

_logger = logging.getLogger('sd-client')

module_list = {
    'Silent Dune Simple Example': {
        'module': 'SilentDuneSimpleModuleExample',
    },
}


class SilentDuneSimpleModuleExample(BaseModule):

    _t_debug_message = 0  # Timed event property for debug message loop.
    _t_reload_task = 0  # Timed event property for reloading firewall rules.

    _enabled = False

    def __init__(self):

        self._arg_name = 'simple'  # Set argparser name
        self._config_section_name = 'simple_module'  # Set configuration file section name

        # Enable multi-threading
        self.wants_processing_thread = True

    def process_loop(self):

        # Print a debug message every 45 seconds
        if self.timed_event('_t_debug_message', 45):
            _logger.debug('{0} 45 second debug message'.format(self.get_name()))

        # Tell the firewall module to reload the firewall rules every 2 minutes.
        if self.timed_event('_t_reload_task', 120):
            # Notify the firewall module to reload the rules.
            task = QueueTask(TASK_FIREWALL_RELOAD_RULES,
                             src_module=self.get_name(),
                             dest_module=SilentDuneClientFirewallModule().get_name())
            self.send_parent_task(task)



#
# Authors: Ma He <mahe.itsec@gmail.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os

from silentdune_client.modules.firewall.manager.slots import Slots
from silentdune_client.builders import iptables as ipt
from silentdune_client.modules.automation.auto_discovery.base_discovery_service import BaseDiscoveryService
from silentdune_client.modules.firewall.manager.iptables_utils import create_iptables_icmp_all

_logger = logging.getLogger('sd-client')


class IcmpDiscovery(BaseDiscoveryService):
    """
    Auto discover ICMP.
    """

    _slot = Slots.icmp
    _config_property_name = '_disable_auto_icmp'

    def _discover_iptables(self):

        rules = list()

        _logger.debug('{0}: adding ICMP Rules.'.format(self.get_name()))
        rules.append(create_iptables_icmp_all(self._slot, transport=ipt.TRANSPORT_IPV4))
        rules.append(create_iptables_icmp_all(self._slot, transport=ipt.TRANSPORT_IPV6))

        return rules



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from silentdune_client.utils.node_info import NodeInformation


_logger = logging.getLogger('sd-client')


class BaseService(object):
    """
    Simple base object for auto service.
    """

    _slot = 0  # Set this to the proper slot level value.
    _service_name = 'unknown'  # Config property name in SilentDuneAutoServicesModule.
    _module = 'SilentDuneAutoServicesModule'

    def get_name(self):
        """
        :return: module name
        """
        # return self._name
        return type(self).__name__

    def get_slot(self):
        return self._slot

    def get_service_name(self):
        return self._service_name

    def discover(self):
        """
        Please DO NOT override this method. Override the platform specific discover methods.
        """
        if NodeInformation.firewall_platform == 'iptables':
            return self._discover_iptables(), self._slot

    def _discover_iptables(self):
        """
        Override this method to discover external services and return either a single machine_subset object or a
        list of machine_subset objects to add to the firewall rules.

        See: modules/firewall/manage/iptables_utils.py for iptables rule builder methods.

        :return: Single iptables machine_subset object or list of objects.
        """

        pass


#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Generic utilities that apply to all firewall platforms.
"""

from silentdune_client.modules.firewall.manager.iptables_utils import create_iptables_generic_dns_egress_rules, \
    create_iptables_generic_ntp_egress_rules
from silentdune_client.utils.node_info import NodeInformation


def create_generic_dns_egress_rules():
    """
    Create generic DNS access rules.  Usually used temporarily to all DNS queries to succeed.
    :return: rules.
    """

    if NodeInformation.firewall_platform == 'iptables':
        return create_iptables_generic_dns_egress_rules()
    else:
        raise AttributeError


def create_generic_ntp_egress_rules():
    """
    Create generic NTP access rules.  Usually used temporarily to all NTP queries to succeed.
    :return: rules.
    """

    if NodeInformation.firewall_platform == 'iptables':
        return create_iptables_generic_ntp_egress_rules()
    else:
        raise AttributeError



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


import logging
import os
import re
import sys
import pkg_resources

from silentdune_client.models.iptables_log_event import IPLogEvent
from silentdune_client.builders import iptables as ipt
from silentdune_client.modules import BaseModule, QueueTask
from silentdune_client.modules.firewall.manager import SilentDuneClientFirewallModule, \
    TASK_FIREWALL_INSERT_RULES

_logger = logging.getLogger('sd-client')

""" Let other modules subscribe to logging events """
TASK_SUBSCRIBE_LOG_EVENTS = 100
TASK_UNSUBSCRIBE_LOG_EVENTS = 110
TASK_LOG_EVENT = 9800  # Line up with the slot id of log records.

# Define the available Module classes.
module_list = {
    'Silent Dune Firewall Manager': {
        'module': 'SilentDuneClientLoggingModule',
    },
}


class SilentDuneClientLoggingModule(BaseModule):
    """ Silent Dune Logging Module """

    priority = 10  # Second highest loading priority.

    _log_file_list = ['/var/log/messages', '/var/log/kern.log']
    _log_file = None
    _log_handle = None  # Log file handle.
    _log_offset = 0  # Current offset in file.
    _terminators = ('\r\n', '\n', '\r')

    _subscribers = list()

    # Module properties
    _limit_rate = None
    _limit_rate_regex = "^(\d*?)/(second|minute|hour|day|sec|min)$"

    _startup = True

    def __init__(self):

        self._arg_name = 'logging'
        self._config_section_name = 'logging_module'

        # Enable multi-threading
        self.wants_processing_thread = True

        self._enabled = False

        try:
            self._version = pkg_resources.get_distribution(__name__).version
        except:
            self._version = 'unknown'

    def add_installer_arguments(self, parser):
        """
        Virtual Override
        Add our module's argparser arguments
        """

        # Create a argument group for our module
        group = parser.add_argument_group('logging module', 'Silent Dune Logging module')

        group.add_argument('--logging-mod-enable', action='store_true',  # noqa
                           help=_('Enable the logging module'))  # noqa

        group.add_argument('--logging-limit-rate', default='10/min', type=str,  # noqa
                           help=_('Limit log events. (rate[/second|/minute|/hour|/day])'))  # noqa

    def validate_arguments(self, args):
        """
        Virtual Override
        Validate command line arguments and save values to our configuration object.
        :param args: An argparse object.
        """
        # See if we have been enabled or not
        if '--logging-mod-enable' not in sys.argv:
            return True

        if args.logging_mod_enable:
            self._enabled = True
            return True

        # Check for any required arguments here

        # Check logging match rate argument against regex.
        if not re.match(self._limit_rate_regex, args.logging_limit_rate):
            print('sdc-install: argument --logging-limit-rate value is invalid.')
            return False

        self._limit_rate = args.logging_limit_rate

        return True

    def validate_config(self, config):
        """
        Virtual Override
        Validate configuration file arguments and save values to our config object.
        :param config: A ConfigParser object.
        """

        # See if we are enabled or not
        try:
            self._enabled = True if config.get(self._config_section_name, 'enabled').lower() == 'yes' else False
        except:
            _logger.debug('{0} configuration section not found in configuration file.'.format(
                self._config_section_name))
            self._enabled = False

        # Only worry about the rest of the configuration items if we are enabled.
        if self._enabled:

            self._limit_rate = config.get(self._config_section_name, 'logging_limit_rate')
            if not re.match(self._limit_rate_regex, self._limit_rate):
                _logger.warn('{0} limit rate value is invalid, defaulting to "6/min".')
                self._limit_rate = '6/min'

        return True

    def prepare_config(self, config):
        """
        Virtual Override
        Return the configuration file structure. Any new configuration items should be added here.
        Note: The order should be reverse of the expected order in the configuration file.
        """

        config.set(self._config_section_name, 'enabled', 'yes' if self._enabled else 'no')
        config.set(self._config_section_name, 'logging_limit_rate', self._limit_rate)

        config.set_comment(self._config_section_name, self._config_section_name,
                           _('; Silent Dune Logging Module Configuration\n'))  # noqa
        config.set_comment(self._config_section_name, 'logging_limit_rate',
                           _('; Limit log events. (rate[/second|/minute|/hour|/day])\n'))  # noqa

        return True

    def service_startup(self):
        _logger.debug('{0} module startup called'.format(self.get_name()))

        # Check to see if we can open the log file or not.
        result = self.open_log()
        self.close_log()

        return result

    def open_log(self):

        for name in self._log_file_list:
            if os.path.exists(name):
                self._log_file = name

        if not self._log_file:
            return False

        try:
            self._log_handle = open(self._log_file)
            self._log_handle.seek(0, 2)  # Seek end of file
            self._log_offset = self._log_handle.tell()
        except IOError:
            _logger.error('Unable to open "{0}" for reading.'.format(self._log_file))
            return False

        return True

    def close_log(self):

        if self._log_handle:
            self._log_handle.close()
            self._log_handle = None

    def service_shutdown(self):
        _logger.debug('{0} thread shutdown called'.format(self.get_name()))
        self.close_log()

    def process_loop(self):
        # _logger.debug('{0} processing loop called'.format(self.get_name()))

        if self._startup:
            self._startup = False

            task = QueueTask(TASK_FIREWALL_INSERT_RULES,
                             src_module=self.get_name(),
                             dest_module=SilentDuneClientFirewallModule().get_name(),
                             data=self.create_logging_rules())
            self.send_parent_task(task)

        if self._log_handle:

            # See if we have iptables log event data
            data = self._log_handle.readline()

            while data:

                if self.node_info.platform == 'linux' and 'kernel: SDC' in data:

                    # See tests/iptables_logging_test.py for example log events.
                    ev = IPLogEvent(data)

                    for mod_name in self._subscribers:
                        # _logger.debug('{0}: sending {1} log event.'.format(self.get_name(), mod_name))
                        task = QueueTask(TASK_LOG_EVENT,
                                         src_module=self.get_name(),
                                         dest_module=mod_name,
                                         data=ev)
                        self.send_parent_task(task)

                data = self._log_handle.readline()

    def process_task(self, task):

        if task:
            t_id = task.get_task_id()

            # Add a subscriber module to our subscribers list.
            if t_id == TASK_SUBSCRIBE_LOG_EVENTS:
                _logger.debug('{0}: {1}: {2}: subscribed.'.format(
                    self.get_name(), task.get_sender(), 'TASK_SUBSCRIBE_LOG_EVENTS'))

                name = task.get_sender()
                if name not in self._subscribers:
                    self._subscribers.append(name)

                    # If we haven't opened the log file, do so now.
                    if not self._log_handle:
                        self.open_log()

            # Remove a subscriber module from our list.
            if t_id == TASK_UNSUBSCRIBE_LOG_EVENTS:
                _logger.debug('{0}: {1}: {2}: unsubscribed.'.format(
                    self.get_name(), task.get_sender(), 'TASK_SUBSCRIBE_LOG_EVENTS'))

                name = task.get_sender()
                if name in self._subscribers:
                    self._subscribers.pop(name)

                    # Close log file if we have no subscribers.
                    if len(self._subscribers) == 0:
                        self.close_log()

    def create_logging_rules(self):

        rules = list()

        rules.append(ipt.get_machine_subset(
            u'ipv4 logging rule',
            9800,
            [
                ipt.get_chain(
                    u'filter',
                    [
                        ipt.get_ring(
                            u'output',
                            ipt.TRANSPORT_IPV4,
                            [
                                ipt.get_rule(
                                    matches=[
                                        ipt.get_match('limit', options=[
                                            ipt.get_match_option('--limit', self._limit_rate),
                                        ]),
                                    ],
                                    jump=ipt.get_jump(target=u'LOG', params=[
                                        ipt.get_jump_option(u'--log-level', value="alert"),
                                        ipt.get_jump_option(u'--log-prefix', value=u'SDC: '),
                                    ])
                                )
                            ]
                        ),
                    ]
                )
            ]
        )
        )

        rules.append(ipt.get_machine_subset(
            u'ipv4 logging rule',
            9800,
            [
                ipt.get_chain(
                    u'filter',
                    [
                        ipt.get_ring(
                            u'output',
                            ipt.TRANSPORT_IPV6,
                            [
                                ipt.get_rule(
                                    matches=[
                                        ipt.get_match('limit', options=[
                                            ipt.get_match_option('--limit', self._limit_rate),
                                        ]),
                                    ],
                                    jump=ipt.get_jump(target=u'LOG', params=[
                                        ipt.get_jump_option(u'--log-level', value="alert"),
                                        ipt.get_jump_option(u'--log-prefix', value=u'SDC: '),
                                    ])
                                )
                            ]
                        ),
                    ]
                )
            ]
        )
        )

        return rules



#
# Authors: Ma He <mahe.itsec@gmail.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import subprocess
import shlex

from silentdune_client.modules.firewall.manager.slots import Slots
from silentdune_client.builders import iptables as ipt
from silentdune_client.modules.automation.auto_discovery.base_discovery_service import BaseDiscoveryService
from silentdune_client.modules.firewall.manager.iptables_utils import create_iptables_udp_egress_ingress_rule
from silentdune_client.utils.misc import which, is_service_running

_logger = logging.getLogger('sd-client')


class NetworkTimeProtocolDiscovery(BaseDiscoveryService):
    """
    Auto discover NTP.
    """

    _slot = Slots.ntp
    _config_property_name = '_disable_auto_ntp'

    def _discover_iptables(self):

        rules = list()

        ntpq = which(u'ntpq')
        if not ntpq:
            _logger.debug('Failed to find program path for "{0}"'.format('ntpq'))
            return rules

        # Check to see if ntpd is running
        if not is_service_running('ntpd'):
            _logger.debug('ntpd is not running.')
            return rules

        p = subprocess.Popen(shlex.split('ntpq -p -n'), stdout=subprocess.PIPE)
        stdoutdata, stderrdata = p.communicate()
        result = p.wait()
        
        if stderrdata is None:
            data = stdoutdata.decode('utf-8')
            for line in data.split('\n'):
                item = line.split(' ', 1)
                if item[0][:1] == '+' or item[0][:1] == '-' or item[0][:1] == '*' or item[0][:1] == 'x' or \
                                item[0][:1] == '.' or item[0][:1] == '#' or item[0][:1] == 'o':
                    ipaddr = item[0][1:]

                    _logger.debug('{0}: adding NTP Client Rules for {1}'.format(self.get_name(), ipaddr))
                    rules.append(create_iptables_udp_egress_ingress_rule(
                        ipaddr, 123, self._slot, transport=ipt.TRANSPORT_AUTO))

        return rules



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import subprocess

from silentdune_client.modules.automation.auto_services.base_service import BaseService
from silentdune_client.utils.misc import which, is_service_running

_logger = logging.getLogger('sd-client')

DOCKERCOMMANDS = {
    'ps': 'docker ps -a --format="{{.ID}}|{{.Image}}|{{.CreatedAt}}|{{.RunningFor}}|{{.Status}}"',
    'inspect': 'docker inspect {0}',
}


class DockerService(BaseService):
    _service_name = '_service_docker'

    def _discover_iptables(self):
        """
        Look for running docker service. If found, check for containers that require firewall rules.
        :return:
        """

        rules = list()

        docker = which('docker')
        if not docker:
            _logger.debug("{0}: Failed to find 'docker' executable.".format(self._module))
            return rules

        if not is_service_running('docker'):
            _logger.debug("{0}: Docker service not running.".format(self._module))
            return rules

        p = subprocess.Popen([u'ntpq', u'-p', u'-n'], stdout=subprocess.PIPE)
        stdoutdata, stderrdata = p.communicate()
        result = p.wait()

        if stderrdata is None:
            data = stdoutdata.decode('utf-8')
            for line in data.split('\n'):
                items = line.split('|')

                #
                # service fields:
                #   node_id
                #   pid char(100)
                #   name char(250)
                #   type char(100) (system service, docker container, ...)
                #   running bool
                #   runningfor char(100)
                #   panic bool
                #   delete bool (remove service record on purge)
                #
                #
                #



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from enum import IntEnum
import logging
import requests
import socket


from silentdune_client.models.iptables_rules import IPMachineSubset, Bundle
from silentdune_client.models.node import Node, NodeBundle
from silentdune_client.models.global_preferences import GlobalPreferences
from silentdune_client.utils.console import ConsoleBase

_logger = logging.getLogger('sd-client')


class ConnStatus(IntEnum):

    unknown = 1
    authenticated = 2
    not_registered = 3


class SilentDuneConnection (ConsoleBase):

    # Server Authentication
    _oauth_crypt_token = None  # Fernet Encrypted oauth2 information
    _machine_id = None  # Machine id to use for authentication
    _cookies = None

    status = ConnStatus.unknown

    # Connection Information
    _server = None
    _base_url = None
    _no_tls = False
    _port = -1
    _user = None
    _password = None
    _ip = None

    def __init__(self, server, no_tls, port):

        self._server = server
        self._no_tls = no_tls
        self._port = port

    def _build_base_url(self):

        # Build base URL
        self._base_url = 'http://' if self._no_tls else 'https://'
        self._base_url += self._server
        self._base_url += '' if self._port == -1 else ':{0}'.format(self._port)

    def _set_authentication(self):
        """
        Set the server Authentication information for each request.  During the install process, oauth2 authentication
        is used in a cookie. While the service is running, machine_id authentication is used in a request header.
        :return:
        """

        if self._machine_id:
            return {'node': self._machine_id}, None

        if self._oauth_crypt_token:
            return None, dict(token=self._oauth_crypt_token)

        return None, None

    def _make_json_request(self, reqtype, url, data=None):

        reply = None
        status_code = None
        rq = None

        if not self.status == ConnStatus.authenticated:
            _logger.error('Not authenticated to SD server.')
            return reply, status_code, rq

        try:

            self._build_base_url()
            u = '{0}{1}'.format(self._base_url, url)

            # Set the server authentication method.
            headers, cookies = self._set_authentication()

            if reqtype is 'GET':
                rq = requests.get(u, cookies=cookies, headers=headers, timeout=30)
            elif reqtype is 'POST':
                rq = requests.post(u, data=data, cookies=cookies, headers=headers, timeout=30)
            elif reqtype is 'PUT':
                rq = requests.put(u, data=data, cookies=cookies, headers=headers, timeout=30)
            elif reqtype is 'DELETE':
                rq = requests.delete(u, cookies=cookies, headers=headers, timeout=30)
            elif reqtype is 'HEAD':
                rq = requests.head(u, cookies=cookies, headers=headers, timeout=30)
            elif reqtype is 'OPTIONS':
                rq = requests.options(u, cookies=cookies, headers=headers, timeout=30)

        except requests.Timeout:
            _logger.error('Server request timeout.')
        except requests.ConnectionError:
            _logger.error('Server connection error.')
        except requests.RequestException:
            _logger.error('Server request failed.')
        else:

            try:
                reply = rq.json()
            except ValueError:
                pass

            status_code = rq.status_code

        return reply, status_code, rq

    # The purpose of this method is to authenticate the user and password against the SD server and
    # retrieve the encrypted Oauth2 token.
    def connect_with_password(self, username, password):

        if not username or not password:
            _logger.error('Invalid parameter passed.')
            raise ValueError

        self.status = ConnStatus.unknown

        self.cwrite('Resolving server...  ')

        try:
            self._ip = socket.gethostbyname(self._server)
        except socket.error:
            _logger.error('Unable to resolve server ({0})'.format(self._server))
            return False

        self.cwriteline('[OK]', 'Server successfully resolved.')

        self.cwrite('Attempting to authenticate with SD server...  ')

        # Make a GET request so we can get the CSRF token.
        try:

            self._build_base_url()
            resp = requests.get('{0}/accounts/login/'.format(self._base_url))
            _logger.debug('{0}/accounts/login/'.format(self._base_url))

            if resp.status_code != requests.codes.ok:
                _logger.error('Unable to retrieve CSRF token ({0})'.format(resp.status_code))
                return False

            csrf = resp.cookies['csrftoken']

        except Exception:
            _logger.error('CSRF token request attempt failed.')
            return False

        try:

            # Make a POST authentication request to get the encrypted oauth2 token
            resp = requests.post('{0}/accounts/login/'.format(self._base_url),
                                 cookies=resp.cookies,
                                 data={'grant_type': 'password', 'username': username, 'password': password,
                                 'csrfmiddlewaretoken': csrf})

            if resp.status_code != requests.codes.ok:
                _logger.error('Unable to authenticate to server ({0})'.format(resp.status_code))
                return False

        except requests.RequestException:
            _logger.error('Authentication request attempt failed')
            return False

        if resp.json() is None:
            _logger.error('Unknown error occurred parsing server response.')

        # Convert reply into JSON
        reply = resp.json()

        # Check reply status value
        if reply['status'] != 'OK':
            _logger.error('Server authentication request failed.')
            return False

        # Save token and cookies for later use
        self._oauth_crypt_token = resp.cookies['token']
        self._cookies = resp.cookies

        self.cwriteline('[OK]', 'Successfully authenticated with server.')

        self.status = ConnStatus.authenticated

        return True

    def connect_with_machine_id(self, machine_id):
        """
        Authenticate to the server with node machine_id value.
        :return: True if connected and authenticate, otherwise false.
        """

        self.status = ConnStatus.unknown
        self._machine_id = machine_id

        try:
            self._ip = socket.gethostbyname(self._server)
        except socket.error:
            _logger.error('Unable to resolve server ({0})'.format(self._server))
            return False

        _logger.info('Attempting to authenticate with SD server.')

        try:

            self._build_base_url()
            # Reply contains dict array of Node records.  Reply array should be empty or contain one Node record.
            url = '{0}/api/nodes/?machine_id={1}'.format(self._base_url, self._machine_id)

            # Set the server authentication method.
            headers, cookies = self._set_authentication()

            resp = requests.get(url, cookies=cookies, headers=headers, timeout=30)

            if resp.status_code == requests.codes.ok:

                reply = resp.json()

                if reply is not None and reply[0]['machine_id'] == self._machine_id:
                    self.status = ConnStatus.authenticated
                    _logger.info('successfully authenticated with server.')
                    self.status = ConnStatus.authenticated
                    return True

            elif resp.status_code == requests.codes.forbidden:
                _logger.error('*** Node auth failed, we are not registered on the server.        ***')
                _logger.error('*** This node needs to be registered with the Silent Dune server. ***')
                _logger.error('*** Please run sdc-install to register.                           ***')
                self.status = ConnStatus.not_registered
            else:
                _logger.error('node auth failed, unknown code: {0}'.format(resp.status_code))

        except requests.RequestException:
            _logger.error('node authentication request attempt failed')

        return False

    def get_node_by_machine_id(self, machine_id):
        """
        Request Node object from server filtered by machine_id value.
        :param machine_id:
        :return Node object:
        """

        url = '/api/nodes/?machine_id={0}'.format(machine_id)

        # Reply contains dict array of Node records.  Reply array should be empty or contain one Node record.
        reply, status_code, rq = self._make_json_request('GET', url)

        # If we have a good code and no data, then the node has not been registered yet.
        if status_code == requests.codes.ok:
            if reply is not None and len(reply) != 0:
                return Node(reply[0]), status_code

        _logger.error('Node lookup request failed.')
        return None, status_code

    def get_global_preferences(self):
        """
        Request GlobalPreferences object from server.
        :return Node object:
        """

        url = '/api/globals/'

        # Reply contains dict array of Node records.  Reply array should be empty or contain one Node record.
        reply, status_code, rq = self._make_json_request('GET', url)

        if status_code == requests.codes.ok:
            if reply is not None and len(reply) != 0:
                return GlobalPreferences(reply[0]), status_code

        _logger.error('Global preferences lookup request failed.')
        return None, status_code

    def iptables_send_alert(self, log):
        """
        Send an IPTables system log event to the server.
        :param data: IPLogEvent object
        :return: id
        """
        child_log = None

        if log.icmp_error_header:
            child_log = log.icmp_error_header
            log.icmp_error_header = None
            log_d = log.to_dict()
        else:
            log_d = log.to_dict()

        try:
            reply, status_code, rq = self._make_json_request('POST', '/api/alerts/iptables/', log_d)

            if status_code == requests.codes.created:
                if not child_log:
                    return reply['id'], status_code

                # If we have a child record, set the parent_id and send it.
                child_log.parent_id = reply['id']
                child_d = child_log.to_dict()

                reply, status_code, rq = self._make_json_request('POST', '/api/alerts/iptables/', child_d)

                if status_code == requests.codes.created:
                    return reply['id'], status_code
        except:
            pass

        return None, requests.codes.teapot

    def register_node(self, node):
        """
        Register Node on server.
        :param node:
        :return Node:
        """
        try:
            node_d = node.to_dict()
        except AttributeError:
            _logger.critical('Node parameter is not valid in register_node method.')
            return None, requests.codes.teapot

        try:
            # Reply contains single Node record
            reply, status_code, rq = self._make_json_request('POST', '/api/nodes/', node_d)

            if reply is not None and status_code is not None:

                # 201 means the node record was created successfully.
                if status_code == requests.codes.created or status_code == requests.codes.ok:
                    return Node(reply), status_code

            _logger.debug('Register node failed ({0}).'.format(status_code))
            return None, status_code

        except:
            pass

        return None, requests.codes.teapot

    def update_node(self, node_id, data):
        """
        Update Node information on server.  Currently only "active" and "sync" fields should be updated.
        :param data: Dictionary with values for "active" and/or "sync" node fields.
        :return Node:
        """
        try:
            # Reply contains single Node record
            reply, status_code, rq = self._make_json_request('PUT', '/api/nodes/{0}/'.format(node_id), data)

            # _logger.debug('Node Update Request: {0}: {1}'.format(status_code, reply))

            if reply is not None and status_code is not None:

                # 200 means the node record was update successfully.
                if status_code == requests.codes.ok:
                    return Node(reply), status_code

            _logger.debug('Update node failed ({0}).'.format(status_code))
            return None, status_code

        except:
            # _logger.debug('ex: {0}: {1}'.format(status_code, reply))
            pass

        return None, requests.codes.teapot

    def get_node_bundle_by_node_id(self, node_id):
        """
        Get the Node Bundle from the server using the node id
        :param node_id:
        :return:
        """
        url = '/api/nodes/{0}/bundle/'.format(node_id)

        reply, status_code, rq = self._make_json_request('GET', url)

        if reply is not None and status_code == requests.codes.ok and len(reply) > 0:
            return NodeBundle(reply[0]), status_code

        _logger.error('Failed to retrieve node bundle by node id ({0}).'.format(node_id))
        return None, status_code

    def get_bundle_by_name(self, name):
        """
        Request Bundle object from server filtered by name value.
        :param name:
        :return Bundle:
        """
        url = '/api/bundles/?name={0}'.format(name)

        reply, status_code, rq = self._make_json_request('GET', url)

        if reply is not None and status_code == requests.codes.ok and len(reply) > 0:
            return Bundle(reply[0]), status_code

        _logger.error('Failed to retrieve bundle by name ({0}).'.format(name))
        return None, status_code

    def get_bundle_by_id(self, bundle_id):
        """
        Request Bundle object from server filtered by name value.
        :param name:
        :return Bundle:
        """
        url = '/api/bundles/{0}/'.format(bundle_id)

        reply, status_code, rq = self._make_json_request('GET', url)

        if reply is not None and status_code == requests.codes.ok and len(reply) > 0:
            return Bundle(reply), status_code

        _logger.error('Failed to retrieve bundle by id ({0}).'.format(bundle_id))
        return None, status_code

    def get_default_bundle(self):
        """
        Request default Bundle object from server.
        :return Bundle:
        """

        url = '/api/bundles/?default'

        # Reply contains single Bundle record
        reply, status_code, rq = self._make_json_request('GET', url)

        if reply is not None and status_code == requests.codes.ok and len(reply) > 0:
            return Bundle(reply[0]), status_code

        return None, status_code

    def create_node_bundle(self, nb):

        if nb is None:
            _logger.critical('NodeBundle parameter is not valid in create_node_bundle method.')
            return None, requests.codes.teapot

        url = '/api/nodes/{0}/bundle/'.format(nb.node)

        # NodeBundleViewSet will update or create with a POST request.
        reply, status_code, rq = self._make_json_request('POST', url, nb.to_dict())

        if reply is not None and \
                (status_code == requests.codes.created or status_code == requests.codes.ok):
            return NodeBundle(reply), status_code

        return None, status_code

    def update_node_bundle(self, nb):

        if nb is None:
            _logger.critical('NodeBundle parameter is not valid in update_node_bundle method.')
            return None, requests.codes.teapot

        url = '/api/nodes/{0}/bundle/{1}/'.format(nb.node, nb.bundle)

        # NodeBundleViewSet will update or create with a POST request.
        reply, status_code, rq = self._make_json_request('PUT', url, nb.to_dict())

        if reply is not None and \
                (status_code == requests.codes.created or status_code == requests.codes.ok):
            return NodeBundle(reply), status_code

        return None, status_code

    def get_bundle_machine_subsets(self, nb):

        if not isinstance(nb, NodeBundle):
            _logger.critical('NodeBundle parameter is not valid in create_or_update_node_bundle method.')
            return None, requests.codes.teapot

        url = '/api/bundles/{0}/machine_subsets/'.format(nb.bundle)

        # Reply contains dict array of IPMachineSet records.
        reply, status_code, rq = self._make_json_request('GET', url)

        ol = list()

        if reply is not None and status_code == requests.codes.ok:

            # Iterate over the reply to get our machine subset records.
            for o in reply:

                url = '/api/iptables/machine_subsets/{0}/'.format(o['machine_subset'])

                # Reply contains single IPMachineSubset record.
                reply, status_code, rq = self._make_json_request('GET', url)

                if reply is not None and status_code == requests.codes.ok:
                    ol.append(IPMachineSubset(reply))

        if len(ol) == 0:
            return None, status_code

        return ol, status_code

    # def write_bundle_to_file(self, path, bundle):
    #     """
    #     Write the bundle machine set data to a file.
    #     This process starts by looping through each of the iptable table names, then calling o.write_chains() to
    #     setup the custom iptables chain names for that table.  Then o.write() is called to write out the rules for
    #     that table. Chain names are defined by the built-in chain name, slot number and pk id of the IPRing object.
    #     """
    #
    #     files = list()
    #
    #     if bundle is None:
    #         _logger.error('Array of chainset objects not valid.')
    #         return files
    #
    #     # Write out machine sets to individual files with the slot number as the name.
    #     for ms in bundle:
    #         file = os.path.join(path, u'{0}.bundle'.format(ms.slot))
    #
    #         _logger.debug('Writting set {0} rules to -> {1}'.format(ms.slot, file))
    #         with open(file, 'w') as handle:
    #             handle.write('{0}\n'.format(ms.to_json()))
    #
    #     for v in {u'ipv4', u'ipv6'}:
    #
    #         file = os.path.join(path, u'{0}.rules'.format(v))
    #
    #         _logger.debug('Writting {0} rules to -> {1}'.format(v, file))
    #
    #         files.append(file)
    #
    #         writer = IPRulesFileWriter(bundle)
    #         writer.write_to_file(file, v)
    #
    #     return files



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Generic Firewall function that apply to iptables platform.
"""

import socket

from silentdune_client.builders import iptables as ipt
from silentdune_client.modules.firewall.manager.slots import Slots
from silentdune_client.utils.misc import is_valid_ipv4_address, is_valid_ipv6_address


# Note: If this method is put in manager.utils it creates a circular reference.
def check_transport_value(ipaddr, transport):
    """
    Check the given transport version against the given IP address.
    :param ipaddr: IP address to check.
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :return: The correct ipt transport version.
    """

    if transport == ipt.TRANSPORT_AUTO:
        if is_valid_ipv6_address(ipaddr):
            transport = ipt.TRANSPORT_IPV6
        elif is_valid_ipv4_address(ipaddr):
            transport = ipt.TRANSPORT_IPV4
        else:
            raise ValueError('Invalid transport version ({0})'.format(transport))
    elif transport == ipt.TRANSPORT_IPV4:
        if not is_valid_ipv4_address(ipaddr):
            raise ValueError('Invalid transport version ({0})'.format(transport))
    elif transport == ipt.TRANSPORT_IPV6:
        if not is_valid_ipv6_address(ipaddr):
            raise ValueError('Invalid transport version ({0})'.format(transport))
    else:
        raise ValueError('Invalid transport version ({0})'.format(transport))

    return transport


def resolve_hostname(name, transport):
    """
    Resolve a hostname to an IP address.
    :param name: Hostname to resolve.
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :return: IP address. Using ipt.TRANSPORT_AUTO will always return an IPv4 IP address.
    """

    if not name:
        raise ValueError('Invalid host name parameter')

    if transport == ipt.TRANSPORT_IPV6:
        addr = socket.getaddrinfo(name, None, socket.AF_INET6, 0, socket.IPPROTO_TCP)[0][4][0]
    else:
        addr = socket.getaddrinfo(name, None, socket.AF_INET, 0, socket.IPPROTO_TCP)[0][4][0]

    return addr


def create_iptables_loopback(slot, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create an rule for loopback interface.
    :param slot: Firewall slot number
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """

    if transport == ipt.TRANSPORT_AUTO:
        raise ValueError('Unable to determine transport for loopback')

    return ipt.get_machine_subset(
        desc,
        slot,
        [
            ipt.get_chain(
                u'filter',
                [
                    ipt.get_ring(
                        u'input',
                        transport,
                        [
                            ipt.get_rule(ifacein_name=u'lo', jump=ipt.get_jump(target=u'ACCEPT'))
                        ]
                    ),
                    ipt.get_ring(
                        u'output',
                        transport,
                        [
                            ipt.get_rule(ifaceout_name=u'lo', jump=ipt.get_jump(target=u'ACCEPT'))
                        ]
                    ),
                ]
            )
        ]
    )


def create_iptables_icmp_all(slot, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create an rule for icmp traffic.
    :param slot: Firewall slot number
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """

    if transport == ipt.TRANSPORT_AUTO:
        raise ValueError('Unable to determine transport for icmp')
    elif transport == ipt.TRANSPORT_IPV4:
        protocol = u'icmp'
    elif transport == ipt.TRANSPORT_IPV6:
        protocol = u'icmpv6'

    return ipt.get_machine_subset(
        desc,
        slot,
        [
            ipt.get_chain(
                u'filter',
                [
                    ipt.get_ring(
                        u'input',
                        transport,
                        [
                            ipt.get_rule(
                                ip_protocol_name=protocol,
                                jump=ipt.get_jump(target=u'ACCEPT'))
                        ]
                    ),
                    ipt.get_ring(
                        u'output',
                        transport,
                        [
                            ipt.get_rule(
                                ip_protocol_name=protocol,
                                jump=ipt.get_jump(target=u'ACCEPT'))
                        ]
                    ),
                ]
            )
        ]
    )


def create_iptables_generic_dns_egress_rules():
    """
    Create generic egress rules to allow access to any DNS server.
    :return:
    """

    rules = list()

    rules.append(create_iptables_tcp_egress_ingress_rule(None, 53, Slots.dns, transport=ipt.TRANSPORT_IPV4,
                                                         desc=u'all DNS Rule'))
    rules.append(create_iptables_tcp_egress_ingress_rule(None, 53, Slots.dns, transport=ipt.TRANSPORT_IPV6,
                                                         desc=u'all DNS Rule'))
    rules.append(create_iptables_udp_egress_ingress_rule(None, 53, Slots.dns, transport=ipt.TRANSPORT_IPV4,
                                                         desc=u'all DNS Rule'))
    rules.append(create_iptables_udp_egress_ingress_rule(None, 53, Slots.dns, transport=ipt.TRANSPORT_IPV6,
                                                         desc=u'all DNS Rule'))

    return rules


def create_iptables_generic_ntp_egress_rules():
    """
    Create generic egress rules to allow access to any NTP server.
    :return:
    """

    rules = list()

    rules.append(create_iptables_udp_egress_ingress_rule(None, 123, Slots.ntp, transport=ipt.TRANSPORT_IPV4,
                                                         desc=u'all NTP Rule'))
    rules.append(create_iptables_udp_egress_ingress_rule(None, 123, Slots.ntp, transport=ipt.TRANSPORT_IPV6,
                                                         desc=u'all NTP Rule'))

    return rules


def create_iptables_ingress_rule_dest(ipaddr, port, protocol, slot, state, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create an ingress firewall rule for destination.
    :param ipaddr: Destination IP address, not host name.
    :param port: Destination port to use for connection
    :param protocol: protocol to use, IE: tcp, udp, icmp, ...
    :param slot: Firewall slot number
    :param state: state of iptables: NEW,ESTABLISHED,RELATED
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """

    if ipaddr:
        # Check transport version.
        transport = check_transport_value(ipaddr, transport)
    else:
        if transport == ipt.TRANSPORT_AUTO:
            raise ValueError('Unable to determine transport from ipaddr')

    if port:
        if port < 1 or port > 65536:
            raise ValueError('Invalid port value ({0}).'.format(port))

    if protocol not in [u'tcp', u'udp', u'icmp', u'udplite', u'esp', u'ah', u'sctp']:
        raise ValueError('Invalid protocol ({0}).'.format(protocol))

    if state:
        if state.strip().replace(u'NEW', '').replace(u'ESTABLISHED', '').replace(u'RELATED', '').replace(u',', ''):
            raise ValueError('Invalid Output State ({0}).'.state)
    else:
        raise ValueError('Output State is none.')

    if port:
        matches = [ipt.get_match(u'state', [ipt.get_jump_option(u'--state', state), ], ),
                   ipt.get_match(protocol, [ipt.get_match_option(u'--dport', port), ], ), ]
    else:
        matches = [ipt.get_match(u'state', [ipt.get_jump_option(u'--state', state), ], ), ]

    if ipaddr:
        rule = ipt.get_rule(ip_protocol_name=protocol, dest_address=ipaddr, matches=matches,
                            jump=ipt.get_jump(target=u'ACCEPT'))
    else:
        rule = ipt.get_rule(ip_protocol_name=protocol, matches=matches, jump=ipt.get_jump(target=u'ACCEPT'))

    return ipt.get_machine_subset(
        desc,
        slot,
        [
            ipt.get_chain(
                u'filter',
                [
                    ipt.get_ring(
                        u'input',
                        transport,
                        [
                            rule
                        ]
                    ),
                ]
            )
        ]
    )


def create_iptables_ingress_rule_source(ipaddr, port, protocol, slot, state, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create an ingress firewall rule for source.
    :param ipaddr: Source IP address, not host name.
    :param port: Source port to use for connection
    :param protocol: protocol to use, IE: tcp, udp, icmp, ...
    :param slot: Firewall slot number
    :param state: state of iptables: NEW,ESTABLISHED,RELATED
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """

    if ipaddr:
        # Check transport version.
        transport = check_transport_value(ipaddr, transport)
    else:
        if transport == ipt.TRANSPORT_AUTO:
            raise ValueError('Unable to determine transport from ipaddr')

    if port:
        if port < 1 or port > 65536:
            raise ValueError('Invalid port value ({0}).'.format(port))

    if protocol not in [u'tcp', u'udp', u'icmp', u'udplite', u'esp', u'ah', u'sctp']:
        raise ValueError('Invalid protocol ({0}).'.format(protocol))

    if state:
        if state.strip().replace(u'NEW', '').replace(u'ESTABLISHED', '').replace(u'RELATED', '').replace(u',', ''):
            raise ValueError('Invalid Output State ({0}).'.state)
    else:
        raise ValueError('Output State is none.')

    if port:
        matches = [ipt.get_match(u'state', [ipt.get_jump_option(u'--state', state), ], ),
                   ipt.get_match(protocol, [ipt.get_match_option(u'--sport', port), ], ), ]
    else:
        matches = [ipt.get_match(u'state', [ipt.get_jump_option(u'--state', state), ], ), ]

    if ipaddr:
        rule = ipt.get_rule(ip_protocol_name=protocol, source_address=ipaddr, matches=matches,
                            jump=ipt.get_jump(target=u'ACCEPT'))
    else:
        rule = ipt.get_rule(ip_protocol_name=protocol, matches=matches, jump=ipt.get_jump(target=u'ACCEPT'))

    return ipt.get_machine_subset(
        desc,
        slot,
        [
            ipt.get_chain(
                u'filter',
                [
                    ipt.get_ring(
                        u'input',
                        transport,
                        [
                            rule
                        ]
                    ),
                ]
            )
        ]
    )


def create_iptables_egress_rule_dest(ipaddr, port, protocol, slot, state, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create an egress firewall rule for destination.
    :param ipaddr: Destination IP address, not host name.
    :param port: Destination port to use for connection
    :param protocol: protocol to use, IE: tcp, udp, icmp, ...
    :param slot: Firewall slot number
    :param state: state of iptables: NEW,ESTABLISHED,RELATED
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """

    if ipaddr:
        # Check transport version.
        transport = check_transport_value(ipaddr, transport)
    else:
        if transport == ipt.TRANSPORT_AUTO:
            raise ValueError('Unable to determine transport from ipaddr')

    if port:
        if port < 1 or port > 65536:
            raise ValueError('Invalid port value ({0}).'.format(port))

    if protocol not in [u'tcp', u'udp', u'icmp', u'udplite', u'esp', u'ah', u'sctp']:
        raise ValueError('Invalid protocol ({0}).'.format(protocol))

    if state:
        if state.strip().replace(u'NEW', '').replace(u'ESTABLISHED', '').replace(u'RELATED', '').replace(u',', ''):
            raise ValueError('Invalid Output State ({0}).'.state)
    else:
        raise ValueError('Output State is none.')

    if port:
        matches = [ipt.get_match(u'state', [ipt.get_jump_option(u'--state', state), ], ),
                   ipt.get_match(protocol, [ipt.get_match_option(u'--dport', port), ], ), ]
    else:
        matches = [ipt.get_match(u'state', [ipt.get_jump_option(u'--state', state), ], ), ]

    if ipaddr:
        rule = ipt.get_rule(ip_protocol_name=protocol, dest_address=ipaddr, matches=matches,
                            jump=ipt.get_jump(target=u'ACCEPT'))
    else:
        rule = ipt.get_rule(ip_protocol_name=protocol, matches=matches, jump=ipt.get_jump(target=u'ACCEPT'))

    return ipt.get_machine_subset(
        desc,
        slot,
        [
            ipt.get_chain(
                u'filter',
                [
                    ipt.get_ring(
                        u'output',
                        transport,
                        [
                            rule
                        ]
                    ),
                ]
            )
        ]
    )


def create_iptables_egress_rule_source(ipaddr, port, protocol, slot, state, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create an egress firewall rule for source.
    :param ipaddr: Source IP address, not host name.
    :param port: Source port to use for connection
    :param protocol: protocol to use, IE: tcp, udp, icmp, ...
    :param slot: Firewall slot number
    :param state: state of iptables: NEW,ESTABLISHED,RELATED
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """

    if ipaddr:
        # Check transport version.
        transport = check_transport_value(ipaddr, transport)
    else:
        if transport == ipt.TRANSPORT_AUTO:
            raise ValueError('Unable to determine transport from ipaddr')

    if port:
        if port < 1 or port > 65536:
            raise ValueError('Invalid port value ({0}).'.format(port))

    if protocol not in [u'tcp', u'udp', u'icmp', u'udplite', u'esp', u'ah', u'sctp']:
        raise ValueError('Invalid protocol ({0}).'.format(protocol))

    if state:
        if state.strip().replace(u'NEW', '').replace(u'ESTABLISHED', '').replace(u'RELATED', '').replace(u',', ''):
            raise ValueError('Invalid Output State ({0}).'.state)
    else:
        raise ValueError('Output State is none.')

    if port:
        matches = [ipt.get_match(u'state', [ipt.get_jump_option(u'--state', state), ], ),
                   ipt.get_match(protocol, [ipt.get_match_option(u'--sport', port), ], ), ]
    else:
        matches = [ipt.get_match(u'state', [ipt.get_jump_option(u'--state', state), ], ), ]

    if ipaddr:
        rule = ipt.get_rule(ip_protocol_name=protocol, source_address=ipaddr, matches=matches,
                            jump=ipt.get_jump(target=u'ACCEPT'))
    else:
        rule = ipt.get_rule(ip_protocol_name=protocol, matches=matches, jump=ipt.get_jump(target=u'ACCEPT'))

    return ipt.get_machine_subset(
        desc,
        slot,
        [
            ipt.get_chain(
                u'filter',
                [
                    ipt.get_ring(
                        u'output',
                        transport,
                        [
                            rule
                        ]
                    ),
                ]
            )
        ]
    )


def create_iptables_tcp_egress_ingress_rule(ipaddr, port, slot, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create a TCP rule that allows external and internal access to the given addr and port.
    :param ipaddr: IP address, not host name.
    :param port: port to use for connection
    :param slot: Firewall slot number
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """
    return create_iptables_egress_ingress_rule(ipaddr, port, u'tcp', slot, transport, desc)


def create_iptables_udp_egress_ingress_rule(ipaddr, port, slot, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create a UDP rule that allows external and internal access to the given addr and port.
    :param ipaddr: IP address, not host name.
    :param port: port to use for connection
    :param slot: Firewall slot number
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """
    return create_iptables_egress_ingress_rule(ipaddr, port, u'udp', slot, transport, desc)


def create_iptables_egress_ingress_rule(ipaddr, port, protocol, slot, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create a rule that allows external and internal access to the given addr and port.
    Create a rule for client.
    :param ipaddr: IP address, not host name.
    :param port: port to use for connection
    :param protocol: protocol to use, IE: tcp, udp, icmp, ...
    :param slot: Firewall slot number
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """

    rules = list()

    rules.append(create_iptables_egress_rule_dest(ipaddr, port, protocol, slot, u'NEW,ESTABLISHED', transport, desc))
    rules.append(create_iptables_ingress_rule_source(ipaddr, port, protocol, slot, u'ESTABLISHED', transport, desc))

    return rules


def create_iptables_tcp_ingress_egress_rule(ipaddr, port, slot, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create a TCP rule that allows external and internal access to the given addr and port.
    :param ipaddr: IP address, not host name.
    :param port: port to use for connection
    :param slot: Firewall slot number
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """
    return create_iptables_ingress_egress_rule(ipaddr, port, u'tcp', slot, transport, desc)


def create_iptables_udp_ingress_egress_rule(ipaddr, port, slot, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create a UDP rule that allows external and internal access to the given addr and port.
    :param ipaddr: IP address, not host name.
    :param port: port to use for connection
    :param slot: Firewall slot number
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """
    return create_iptables_ingress_egress_rule(ipaddr, port, u'udp', slot, transport, desc)


def create_iptables_ingress_egress_rule(ipaddr, port, protocol, slot, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create a rule that allows internal and external access to the given addr and port.
    Create a rule for Server.
    :param ipaddr: IP address, not host name.
    :param port: port to use for connection
    :param protocol: protocol to use, IE: tcp, udp, icmp, ...
    :param slot: Firewall slot number
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """

    rules = list()

    rules.append(create_iptables_ingress_rule_dest(ipaddr, port, protocol, slot, u'NEW,ESTABLISHED', transport, desc))
    rules.append(create_iptables_egress_rule_source(ipaddr, port, protocol, slot, u'ESTABLISHED', transport, desc))

    return rules


def create_iptables_ingress_egress_forward_reject_rule(slot, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create an firewall rule for all reject in INPUT, OUTPUT and FORWARD chain by default.
    :param slot: Firewall slot number
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """

    rules = list()

    rules.append(create_iptables_ingress_reject_rule(slot, transport))
    rules.append(create_iptables_egress_reject_rule(slot, transport))
    rules.append(create_iptables_forward_reject_rule(slot, transport))

    return rules


def create_iptables_ingress_egress_reject_rule(slot, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create an firewall rule for all reject in INPUT, OUTPUT chain by default.
    :param slot: Firewall slot number
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """

    rules = list()

    rules.append(create_iptables_ingress_reject_rule(slot, transport))
    rules.append(create_iptables_egress_reject_rule(slot, transport))

    return rules


def create_iptables_ingress_reject_rule(slot, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create an firewall rule for all reject in INPUT chain by default.
    :param slot: Firewall slot number
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """

    if transport == ipt.TRANSPORT_AUTO:
        raise ValueError('Unable to determine transport')

    rule = ipt.get_rule(jump=ipt.get_jump(target=u'REJECT'))

    return ipt.get_machine_subset(
        desc,
        slot,
        [
            ipt.get_chain(
                u'filter',
                [
                    ipt.get_ring(
                        u'input',
                        transport,
                        [
                            rule
                        ]
                    ),
                ]
            )
        ]
    )


def create_iptables_egress_reject_rule(slot, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create an firewall rule for all reject in OUTPUT chain by default.
    :param slot: Firewall slot number
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """

    if transport == ipt.TRANSPORT_AUTO:
        raise ValueError('Unable to determine transport')

    rule = ipt.get_rule(jump=ipt.get_jump(target=u'REJECT'))

    return ipt.get_machine_subset(
        desc,
        slot,
        [
            ipt.get_chain(
                u'filter',
                [
                    ipt.get_ring(
                        u'output',
                        transport,
                        [
                            rule
                        ]
                    ),
                ]
            )
        ]
    )


def create_iptables_forward_reject_rule(slot, transport=ipt.TRANSPORT_AUTO, desc=''):
    """
    Create an firewall rule for all reject in FORWARD chain by default.
    :param slot: Firewall slot number
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """

    if transport == ipt.TRANSPORT_AUTO:
        raise ValueError('Unable to determine transport')

    rule = ipt.get_rule(jump=ipt.get_jump(target=u'REJECT'))

    return ipt.get_machine_subset(
        desc,
        slot,
        [
            ipt.get_chain(
                u'filter',
                [
                    ipt.get_ring(
                        u'forward',
                        transport,
                        [
                            rule
                        ]
                    ),
                ]
            )
        ]
    )


def create_iptables_remote_config(ipaddr, mask, port, protocol, slot, uid=None, gid=None, transport=ipt.TRANSPORT_AUTO,
                                  desc=''):
    """
    Create firewall rules for remote config.
    :param ipaddr: Destination IP address, not host name.
    :param mask: Netmask
    :param port: Destination port to use for connection
    :param protocol: protocol to use, IE: tcp, udp, icmp, ...
    :param slot: Firewall slot number
    :param transport: ipt.TRANSPORT_AUTO, ipt.TRANSPORT_IPV4 or ipt.TRANSPORT_IPV6
    :param desc: Extra description of the rule.
    :return:
    """

    rules = list()

    if ipaddr:
        # Check transport version.
        transport = check_transport_value(ipaddr, transport)
    else:
        if transport == ipt.TRANSPORT_AUTO:
            raise ValueError('Unable to determine transport from ipaddr')

    if mask:
        if transport == ipt.TRANSPORT_IPV4:
            if mask < 1 or mask > 32:
                raise ValueError('Invalid IPV4 mask value ({0}).'.format(mask))
        elif transport == ipt.TRANSPORT_IPV6:
            if mask < 1 or mask > 128:
                raise ValueError('Invalid IPV6 mask value ({0}).'.format(mask))

    if port:
        if port < 1 or port > 65536:
            raise ValueError('Invalid port value ({0}).'.format(port))

    if protocol not in [u'tcp', u'udp', u'icmp', u'udplite', u'esp', u'ah', u'sctp']:
        raise ValueError('Invalid protocol ({0}).'.format(protocol))

    dest_matches = list()
    dest_matches.append(ipt.get_match(u'state', [ipt.get_jump_option(u'--state', u'NEW,ESTABLISHED'), ], ))
    if port:
        dest_matches.append(ipt.get_match(protocol, [ipt.get_match_option(u'--dport', port), ], ))

    jump_options = list()
    if uid:
        jump_options.append(ipt.get_jump_option(u'--uid-owner', uid))

    if gid:
        jump_options.append(ipt.get_jump_option(u'--gid-owner', uid))

    if jump_options:
        dest_matches.append(ipt.get_match(u'owner', jump_options))

    if jump_options:
        rule = ipt.get_rule(ip_protocol_name=protocol, dest_address=ipaddr, dest_mask=mask, matches=dest_matches)
    else:
        rule = ipt.get_rule(ip_protocol_name=protocol, dest_address=ipaddr, dest_mask=mask, matches=dest_matches,
                            jump=ipt.get_jump(target=u'ACCEPT'))

    rules.append(
        ipt.get_machine_subset(
            desc,
            slot,
            [
                ipt.get_chain(
                    u'filter',
                    [
                        ipt.get_ring(
                            u'output',
                            transport,
                            [
                                rule
                            ]
                        ),
                    ]
                )
            ]
        )
    )

    source_matches = list()
    source_matches.append(ipt.get_match(u'state', [ipt.get_jump_option(u'--state', u'ESTABLISHED'), ], ))
    if port:
        source_matches.append(ipt.get_match(protocol, [ipt.get_match_option(u'--sport', port), ], ))

    rule = ipt.get_rule(ip_protocol_name=protocol, source_address=ipaddr, source_mask=mask, matches=source_matches,
                        jump=ipt.get_jump(target=u'ACCEPT'))

    rules.append(
        ipt.get_machine_subset(
            desc,
            slot,
            [
                ipt.get_chain(
                    u'filter',
                    [
                        ipt.get_ring(
                            u'input',
                            transport,
                            [
                                rule
                            ]
                        ),
                    ]
                )
            ]
        )
    )

    return rules



#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
import sys


def setup_logging(logger, debug=False):
    """
    Setup python logging
    """

    # Set our logging options now that we have the program arguments.
    if debug:
        logging.basicConfig(filename=os.devnull,
                            datefmt='%Y-%m-%d %H:%M:%S', level=logging.DEBUG)
        # Setup logging formatter
        formatter = logging.Formatter('%(asctime)s %(levelname)s: %(message)s')
    else:
        logging.basicConfig(filename=os.devnull,
                            datefmt='%Y-%m-%d %H:%M:%S', level=logging.WARNING)
        # Setup logging formatter
        formatter = logging.Formatter('%(levelname)s: %(message)s')

    # Setup stream logging handler
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)

    logger.addHandler(handler)

    # Setup file logging handler
    handler = logging.FileHandler('/var/log/silentdune.log')
    handler.setFormatter(formatter)

    logger.addHandler(handler)


#
# Authors: Robert Abram <robert.abram@entpack.com>
#
# Copyright (C) 2015-2016 EntPack
# see file 'LICENSE' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import argparse
import gettext
import logging
import operator
import os
import shutil
import shlex
import subprocess
import sys

from silentdune_client.modules import __load_modules__
from silentdune_client.utils.configuration import ClientConfiguration
from silentdune_client.utils.console import ConsoleBase
from silentdune_client.utils.log import setup_logging
from silentdune_client.utils.misc import is_service_running, node_info_dump, which
from silentdune_client.utils.node_info import NodeInformation

try:
    from configparser import ConfigParser  # noqa
except ImportError:
    # ver. < 3.0
    from ConfigParser import ConfigParser  # noqa

logging.basicConfig()
_logger = logging.getLogger('sd-client')


class Installer(ConsoleBase):

    # Modules dictionary list
    __modules = None
    __config = None

    # parser args
    args = None
    bad_arg = False

    node_info = None

    previous_firewall_service = None

    # Location of the installed service unit or script file.
    service_out_file = None

    def __init__(self, args, modules, node_info):

        self.__modules = modules
        self.args = args
        self.node_info = node_info
        self.__config = ClientConfiguration()

    def write_config(self):
        """
        Loop through the modules to set their configuration file items and then save the configuration.
        """

        for mod in self.__modules:
            result = mod.prepare_config(self.__config)

            if result is False:  # If prepare_config() returns None, we just want to continue.
                _logger.error('Preparing configuration file items failed in module {0}.'.format(mod.get_name()))
                return False

        # Write the configuration file out.
        return self.__config.write_config()

    def install_service(self):
        """
        Based on everything we know, lets install the init service.
        :return: True if successful, otherwise False.
        """

        self.cwrite('Configuring Silent Dune firewall service...')

        # Figure out our path
        base_path = os.path.split(os.path.realpath(__file__))[0]

        systemd_in_file = os.path.join(base_path, 'init/sdc-firewall.systemd.in')
        init_in_file = os.path.join(base_path, 'init/sdc-firewall.sysv.in')

        # Check and make sure we can find the init scripts.
        if not os.path.exists(systemd_in_file) or \
                not os.path.exists(init_in_file):
            _logger.critical('Unable to find init service files.')
            return False

        firewall_exec = which('sdc-firewall')

        if not firewall_exec:
            self.cwriteline('[Error]', 'Unable to locate our firewall executable.')
            return False

        # Install systemd service file.
        if self.node_info.sysd_installed:

            path = None

            # Determine systemd service unit install directory.
            if os.path.exists('/usr/lib/systemd/system/'):  # Redhat based
                path = '/usr/lib/systemd/system/'
            elif os.path.exists('/lib/systemd/system/'):  # Ubuntu based
                path = '/lib/systemd/system/'
            elif os.path.exists('/etc/systemd/system/'):  # Last resort location
                path = '/etc/systemd/system/'

            if not path:
                self.cwriteline('[Error]', 'Unable to locate systemd service unit path.')
                return False

            self.service_out_file = os.path.join(path, 'sdc-firewall.service')

            # See if we need to copy the service unit file to the destination
            if not os.path.isfile(self.service_out_file):
                shutil.copy(systemd_in_file, self.service_out_file)
                os.chmod(self.service_out_file, 0o644)

        if self.node_info.sysv_installed:

            # http://askubuntu.com/questions/2263/chkconfig-alternative-for-ubuntu-server
            path = '/etc/init.d/'
            self.service_out_file = os.path.join(path, 'sdc-firewall')

            # See if we need to copy the service unit file to the destination
            if not os.path.isfile(self.service_out_file):
                shutil.copy(init_in_file, self.service_out_file)
                os.chmod(self.service_out_file, 0o755)

        # Enable service
        # if not self.node_info.enable_service('sdc-firewall'):
        #    self.cwriteline('[Error]', 'Firewall service failed to enable.')
        #    return False

        # Start service
        # if not self.node_info.start_service('sdc-firewall'):
        #     self.cwriteline('[Error]', 'Firewall service failed to start.')
        #     return False

        self.cwriteline('      [OK]', 'Firewall service installed. Please start "sdc-firewall" service now.')

        return True

    def remove_service(self):

        # Remove the systemd service file.
        if self.node_info.sysd_installed:

            if is_service_running('sdc-firewall'):
                if not self.node_info.stop_service('sdc-firewall'):
                    _logger.debug('Firewall service failed to stop.')

                if not self.node_info.disable_service('sdc-firewall'):
                    _logger.debug('Unable to disable firewall service.')

            if os.path.exists(self.service_out_file):
                os.remove(self.service_out_file)

        if self.node_info.sysv_installed:
            # TODO: Write the sysv service removal code.
            pass

        self.cwriteline('[OK]', 'Firewall service removed.')

    def disable_previous_firewall(self):
        """
        Disable the previous firewall service.
        :return: True if successful, otherwise False.
        """

        if not self.node_info.previous_firewall_service or self.node_info.previous_firewall_service == 'sdc-firewall':
            return True

        # Check to see if the previous firewall service is running.
        if not is_service_running(self.node_info.previous_firewall_service):
            _logger.info('The current firewall service does not seem to be running.')
            return True

        self.cwrite('Stopping the current firewall service...')

        # Stop and Disable the previous firewall service.
        if not self.node_info.stop_service(self.node_info.previous_firewall_service):
            self.cwriteline('[Error]', 'Unable to stop the current firewall service.')
            return False

        self.cwriteline('[OK]', 'Successfully stopped the current firewall service.')

        self.cwrite('Disabling the current firewall service...')

        if not self.node_info.disable_service(self.node_info.previous_firewall_service):
            self.cwriteline('[Error]', 'Unable to disable the current firewall service.')
            return False

        self.cwriteline('[OK]', 'Successfully disabled the current firewall service.')

        return True

    def clean_up(self):
        """
        Use this method to clean up after a failed install
        """
        self.cwrite('Cleaning up...')

        # # The following code can only run if we are running under privileged account.
        # if self.node_info.root_user:
        #
        #     # Remove our firewall service.
        #     self.remove_service()
        #
        #     # Remove the directories the daemon process uses.
        #     self.__config.delete_directories()
        #     _logger.debug('Removed daemon process directories.')
        #
        #     # Remove the system user the daemon process uses.
        #     self.__config.delete_user()
        #     _logger.debug('Removed daemon process system user.')
        #
        #     # Check to see if the previous firewall service is running or not
        #     if self.node_info.previous_firewall_service:
        #         if not is_service_running(self.node_info.previous_firewall_service):
        #             self.node_info.enable_service(self.node_info.previous_firewall_service)
        #             self.node_info.start_service(self.node_info.previous_firewall_service)
        #
        #     # if we are running as root, delete the configuration directory
        #     if self.node_info.config_root is not None \
        #             and os.path.exists(self.node_info.config_root) \
        #             and os.path.realpath(self.node_info.config_root) != '/':
        #         rmdir(self.node_info.config_root)

        self.cwriteline('[OK]', 'Finished cleaning up.')

    def start_install(self):
        """
        Begin installing the Silent Dune Client.
        """

        # If this node is running systemd, remove the pidfile setting.
        if self.node_info.sysd_installed:
            self.__config.delete('settings', 'pidfile')

        # Check to see that the NodeInformation information gathering was successful.
        if self.node_info.error:
            _logger.error('Gathering information about this node failed.')
            return False

        # See if we haven't determined the configuration root directory.
        if not self.node_info.config_root:
            _logger.error('Error determining the configuration root directory.')
            return False

        # if self.node_info.previous_firewall_service == 'sdc-firewall':
        #     self.cwriteline('[Error]', 'Silent Dune firewall service is already running, please uninstall first.')
        #     return False

        #
        # Have each module do their pre install work now.
        #
        for mod in self.__modules:
            result = mod.pre_install()
            if result is not None and result is False:
                return False

        # The following code can only run if we are running under privileged account.
        if self.node_info.root_user:

            # # Create the daemon process user
            # if not self.__config.create_service_user():
            #     return False

            # Create the directories the daemon process uses.
            if not self.__config.create_directories():
                return False

        #
        # Have each module do their install work now.
        #
        for mod in self.__modules:
            result = mod.install_module()
            if result is not None and result is False:
                return False

        # The following code can only run if we are running under privileged account.
        if self.node_info.root_user:

            # Disable the current firewall service
            if self.node_info.previous_firewall_service:
                if not self.disable_previous_firewall():
                    return False

        if not self.write_config():
            return False

        # Have each module do their post install work now.
        for mod in self.__modules:
            result = mod.post_install()
            if result is not None and result is False:
                return False

        # Finally install the firewall service.
        if self.node_info.root_user:

            if not self.install_service():
                return False

            self.node_info.restorecon('/etc/silentdune')

            if not os.path.isdir('/var/run/silentdune'):
                os.makedirs('/var/run/silentdune')

            self.node_info.restorecon('/var/run/silentdune')

        return True


def run():

    # Set global debug value and setup application logging.
    setup_logging(_logger, '--debug' in sys.argv)

    # See if we need to dump information about this node.
    if '--debug' in sys.argv:
        node_info_dump(sys.argv)

    # Get the path where this file is located.
    app_path = os.path.split(os.path.realpath(__file__))[0]
    # Get our package path and package name
    base_path, package_name = os.path.split(app_path)

    # Check and make sure we can find the init scripts.
    if not os.path.exists(os.path.join(app_path, 'init/sdc-firewall.systemd.in')) or \
            not os.path.exists(os.path.join(app_path, 'init/sdc-firewall.sysv.in')):
        print('sdc-install: error: Unable to locate client init scripts, unable to install')
        sys.exit(1)

    # Setup i18n - Good for 2.x and 3.x python.
    kwargs = {}
    if sys.version_info[0] < 3:
        kwargs['unicode'] = True
    gettext.install('sdc_install', **kwargs)

    # Get loadable module list
    module_list = __load_modules__(base_path=base_path)

    # Sort modules by the priority attribute.
    module_list = sorted(module_list, key=operator.attrgetter('priority'))

    # Setup program arguments.
    parser = argparse.ArgumentParser(prog='sdc-install')
    parser.add_argument(_('--debug'), help=_('Enable debug output'), default=False, action='store_true')  # noqa

    # Loop through the module objects and add any argparse arguments.
    for mod in module_list:
        mod.add_installer_arguments(parser)

    args = parser.parse_args()

    # Have each module validate arguments.
    for mod in module_list:
        if mod.validate_arguments(args) is False:  # If return value is None, we just want to continue.
            parser.print_help()
            exit(1)

    node_info = NodeInformation(console_debug=('--debug' in sys.argv))

    # Instantiate the installer object
    i = Installer(args, module_list, node_info)

    # Begin the install process.
    if not i.start_install():

        # Have each module do their uninstall work now.
        for mod in module_list:
            mod.uninstall_module()

        _logger.error('Install aborted.')
        return 1

    return 0


# --- Main Program Call ---
if __name__ == '__main__':
    sys.exit(run())
